/*
 * watchpoint.c:   kernel watchpoint functionality
 *
 * Copyright (C) 2016 Rob Gardner (rob.gardner@oracle.com)
 */


#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <asm/asi.h>
#include <asm/ptrace.h>
#include <asm/pstate.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/watchpoint.h>
#include <asm/adi_64.h>

/*
 * Used to control whether mcd is enabled or not
 *  0           - Not enabled
 *  TSTATE_MCDE - Enabled
 */
unsigned long mcd_enable = 0;
unsigned long mcd_master_enable = 0;

/* does an access touch a watched chunk of memory? */
static int range_overlap(unsigned long watch_addr, unsigned long watch_len,
		unsigned long access_addr, unsigned long access_len)
{
	unsigned long last_store_byte = access_addr + access_len - 1;
	unsigned long last_watch_byte = watch_addr + watch_len - 1;

	if (watch_len == 0 || access_len == 0)
		return 0;
	if (last_store_byte < watch_addr) /* entire store precedes watched region */
		return 0;
	if (last_watch_byte < access_addr) /* entire watched area precedes store */
		return 0;

	return 1;
}

/* small, simple array of watchpoints, searched linearly by find_watchpoint(). */
/* something more elegant might be added later but YAGNI */
watchpoint_t watchpoints[NUM_WATCHPOINTS];
#define MATCH_OVERLAP 0
#define MATCH_EXACT 1

static watchpoint_t *find_watchpoint(unsigned long waddr, int len, int matchtype)
{
	int i;
	watchpoint_t *w;

	for (i=0, w = &watchpoints[0]; i<NUM_WATCHPOINTS; i++, w++) {
		if (matchtype == MATCH_EXACT)
			if (w->addr == waddr && w->len == len)
				return w;
		if (range_overlap(w->addr, w->len, waddr, len))
			return w;
	}
	return NULL;
}

static inline enum direction decode_direction(unsigned int insn)
{
	unsigned long tmp = (insn >> 21) & 1;

	if (!tmp)
		return load;

	tmp = (insn >> 19) & 0x3f; /* op3 */

	switch (tmp) {
	case 0x3c:		/* casa */
	case 0x3e:		/* casxa */
	case 0xf:		/* swap */
	case 0x1f:		/* swapa */
	case 0xd:		/* ldstub */
	case 0x1d:		/* ldstuba */
		return both;
	default:
		return store;
	}
}

static int decode_access_size(unsigned int insn)
{
	unsigned int tmp;

	tmp = ((insn >> 19) & 0xf);
	if (tmp == 11 || tmp == 14) /* ldx/stx */
		return 8;
	tmp &= 3;
	if (!tmp)
		return 4;
	else if (tmp == 3)
		return 16;	/* ldd/std - Although it is actually 8 */
	else if (tmp == 2)
		return 2;
	else {
		return 1;
	}
}

static inline int decode_asi(unsigned int insn, struct pt_regs *regs)
{
	if (insn & 0x800000) {
		if (insn & 0x2000)
			return (unsigned char)(regs->tstate >> 24);	/* %asi */
		else
			return (unsigned char)(insn >> 5);		/* imm_asi */
	} else
		return ASI_P;
}

/* 0x400000 = signed, 0 = unsigned */
static inline int decode_signedness(unsigned int insn)
{
	return (insn & 0x400000);
}

static inline void maybe_flush_windows(unsigned int rs1, unsigned int rs2,
				       unsigned int rd, int from_kernel)
{
	if (rs2 >= 16 || rs1 >= 16 || rd >= 16) {
		if (from_kernel != 0)
			__asm__ __volatile__("flushw");
		else
			flushw_user();
	}
}

static unsigned long fetch_reg(unsigned int reg, struct pt_regs *regs)
{
	unsigned long value;

	if (reg < 16)
		return (!reg ? 0 : regs->u_regs[reg]);
	if (regs->tstate & TSTATE_PRIV) {
		struct reg_window *win;
		win = (struct reg_window *)(regs->u_regs[UREG_FP] + STACK_BIAS);
		value = win->locals[reg - 16];
	} else if (test_thread_flag(TIF_32BIT)) {
		struct reg_window32 __user *win32;
		win32 = (struct reg_window32 __user *)((unsigned long)((u32)regs->u_regs[UREG_FP]));
		get_user(value, &win32->locals[reg - 16]);
	} else {
		struct reg_window __user *win;
		win = (struct reg_window __user *)(regs->u_regs[UREG_FP] + STACK_BIAS);
		get_user(value, &win->locals[reg - 16]);
	}
	return value;
}

static unsigned long *fetch_reg_addr(unsigned int reg, struct pt_regs *regs)
{
	if (reg < 16)
		return &regs->u_regs[reg];
	if (regs->tstate & TSTATE_PRIV) {
		struct reg_window *win;
		win = (struct reg_window *)(regs->u_regs[UREG_FP] + STACK_BIAS);
		return &win->locals[reg - 16];
	} else if (test_thread_flag(TIF_32BIT)) {
		struct reg_window32 *win32;
		win32 = (struct reg_window32 *)((unsigned long)((u32)regs->u_regs[UREG_FP]));
		return (unsigned long *)&win32->locals[reg - 16];
	} else {
		struct reg_window *win;
		win = (struct reg_window *)(regs->u_regs[UREG_FP] + STACK_BIAS);
		return &win->locals[reg - 16];
	}
}

static inline void advance(struct pt_regs *regs)
{
	regs->tpc   = regs->tnpc;
	regs->tnpc += 4;
	if (test_thread_flag(TIF_32BIT)) {
		regs->tpc &= 0xffffffff;
		regs->tnpc &= 0xffffffff;
	}
}

static int read_mcd_tag(unsigned long addr)
{
    int ver;
    long err;
    long faulted;

    __asm__ __volatile__(
	"1:    ldxa [%[addr]] 0x2, %[ver]\n"
	"      mov 0, %[err]\n"
	"2:\n"
	"      .section .fixup,#alloc,#execinstr\n"
	"      .align 4\n"
	"3:     sethi %%hi(2b), %[faulted]\n"
	"       jmpl  %[faulted] + %%lo(2b), %%g0\n"
	"       mov 0xaa, %[err]\n"
	"       .previous\n"
	"       .section __ex_table, \"a\"\n"
	"       .align 4\n"
	"       .word  1b, 3b\n"
	"       .previous\n"
	: [ver] "=r" (ver), [faulted] "=r" (faulted), [err] "=r" (err)
	: [addr] "r"  (addr)
	: "memory"
	);

    if (err == 0)
	return ver;
    else
	return -err;
}

static int set_mcd_tag(unsigned long addr, int ver)
{
    long err;
    long faulted;

    __asm__ __volatile__(
	    "1:    stxa %[ver], [%[addr]] 0x2\n"
	    "      mov 0, %[err]\n"
	    "2:\n"
	    "      .section .fixup,#alloc,#execinstr\n"
	    "      .align 4\n"
	    "3:     sethi %%hi(2b), %[faulted]\n"
	    "       jmpl  %[faulted] + %%lo(2b), %%g0\n"
	    "       mov 0xbb, %[err]\n"
	    "       .previous\n"
	    "       .section __ex_table, \"a\"\n"
	    "       .align 4\n"
	    "       .word  1b, 3b\n"
	    "       .previous\n"
	    : [faulted] "=r" (faulted), [err] "=r" (err)
	    : [ver] "r" (ver), [addr] "r"  (addr)
	    : "memory"
	    );

    if (err)
	    return -err;
    else
	    return ver;
}

/*
 * Set the MCD tags for a range of addresses.
 *
 * If the range spans more than one cache line, the tag must be
 * set for each line
 */
  static int set_mcd_tag_range(unsigned long addr, int len, int ver)
{
	unsigned long base;
	int oldver;

	/*
	 * Check range of address to make sure there are no tags set
	 * anywhere so we don't overlap another watchpoint
	 */
	if (ver) {
		for (base = addr & ~0x3fLL; base < addr+len; base = base + 0x40) {
			if ( (oldver = read_mcd_tag(base)) < 0)
				return -1;
			if (oldver == ver)
				return -1;
		}
	}

	for (base = addr & ~0x3fLL; base < addr+len; base = base + 0x40)
		if (set_mcd_tag(base, ver) < 0)
			return -1;
	return ver;
}

int mcd_not_present(void)
{
#ifdef CONFIG_SPARC_ADI
	return !adi_capable();
#else
	return true;
#endif
}

int set_mcde_state(int enable)
{
	int old_state;
	unsigned long pstate;

	if (mcd_not_present())
		return 0;

	__asm__ __volatile__(
		"rdpr %%pstate, %0\n\t"
		: "=r" (pstate) :: );
	old_state = !!(pstate & PSTATE_MCDE);

	if (enable != old_state) {

		if (enable)
			pstate |= PSTATE_MCDE;
		else
			pstate &= ~PSTATE_MCDE;

		__asm__ __volatile__(
			"wrpr %0, %%g0, %%pstate\n\t"
			: : "r" (pstate) :);
	}

	return old_state;
}
EXPORT_SYMBOL(set_mcde_state);

void disable_mcde(void)
{
	(void) set_mcde_state(0);
}
EXPORT_SYMBOL(disable_mcde);

//int persistent_mcd_debug = MCD_HV_TRACE | MCD_LOAD_TRACE | MCD_STORE_TRACE | MCD_CAS_TRACE;
//int persistent_mcd_debug = MCD_LOAD_TRACE | MCD_STORE_TRACE | MCD_CAS_TRACE;
//int persistent_mcd_debug = MCD_LOAD_TRACE | MCD_STORE_TRACE;
int persistent_mcd_debug = 0;
int mcd_debug;
int mcd_emulate_loads = 0, mcd_emulate_stores = 0, mcd_emulate_ldst = 0;


/*
 * print_record "service"
 */
static void watchpoint_print_record(watchpoint_rec_t *mr)
{
	char symbuf[512];
	unsigned short op3 = (mr->insn >> 19) & 0x3f;

	(void)sprint_symbol(symbuf, mr->tpc);

	switch (mr->dir) {
	case load:
		printk(KERN_ALERT "%8lx\tld%s%d\t[%lx] #%x %%r%d\t= 0x%016lx\t(was: 0x%016lx)\t(%s)\n",
			mr->tpc, mr->sign ? "" : "u",
			mr->access_size, mr->fault_addr, mr->asi, (mr->insn>>25)&0x1f,
			mr->rd_val, mr->old_rd_val, symbuf);
		break;
	case store:
		printk(KERN_ALERT "%8lx\tst%s%d\t[%lx] #%x %%r%d\t= 0x%016lx\t(was: 0x%016lx)\t(%s)\n",
		        mr->tpc, mr->sign ? "" : "u",
			mr->access_size, mr->fault_addr, mr->asi, (mr->insn>>25)&0x1f,
			mr->rd_val, mr->old_mem_val, symbuf);
		break;
	case both:
		if (op3 == 0x3c || op3 == 0x3e) {
			printk(KERN_ALERT "%8lx\tcas%sa\t[%lx] #%x %%r%d, %%r%d\t= 0x%016lx\t(was: 0x%016lx)\t(%s)\n",
				mr->tpc, (op3 == 0x3e) ? "x" : "",
				mr->fault_addr, mr->asi, (mr->insn & 0x1f), (mr->insn>>25)&0x1f,
				mr->rd_val, mr->old_rd_val, symbuf);
		}
		break;
	default:
		break;
	}
}

watchpoint_services_t watchpoint_services =
{
	.print_record = watchpoint_print_record
};


/*
 * Called only from MCD exception handler.
 *
 * This emulates the trapped instruction and checks to see if the
 * accessed address matches one of our watchpoints. If so, then
 * watchpoint callback function is invoked.
 */
int watchpoint(struct pt_regs *regs, unsigned long fault_addr,
				      unsigned long context)
{
	volatile unsigned long *reg_addr, old_val, reg_val;
	static unsigned long timestamp = 0;
	unsigned long now, saved_pstate, addr;

	static int nesting = 0, las_cases = 0;
	unsigned int insn;
	int ret, asi, sign, access_size;

	enum direction dir;
	char symbuf[512];
	watchpoint_rec_t mcd_rec;
	watchpoint_t *w;

	mcd_debug = persistent_mcd_debug;

	if (mcd_not_present())
		return 0;

	/* if no watchpoints set, then do not emulate anything */
	if (find_watchpoint(0, 0xffffffff, MATCH_OVERLAP) == NULL)
		return 0;

	/* when we get here, MCD had better be turned off */
        __asm__ __volatile__(
            "rdpr %%pstate, %0\n\t"
            : "=r" (saved_pstate) : );

	if (saved_pstate & PSTATE_MCDE)
		BUG();

	if (mcd_enable)
		BUG();

	nesting++;
	insn = *(unsigned int *) regs->tpc;
	dir = decode_direction(insn);

	if (timestamp == 0) {
		RDTICK(timestamp);
	}
	RDTICK(now);

	addr = compute_effective_address(regs, insn, 0);
	access_size = decode_access_size(insn);
	asi = decode_asi(insn, regs);
	sign = decode_signedness(insn);

	memset(&mcd_rec, 0, sizeof(watchpoint_rec_t));
	mcd_rec.tpc = regs->tpc;
	mcd_rec.fault_addr = addr;
	mcd_rec.sign = sign;
	mcd_rec.asi = asi;
	mcd_rec.insn = insn;
	mcd_rec.access_size = access_size;
	mcd_rec.dir = dir;
	mcd_rec.elapsed_time = now - timestamp;
	mcd_rec.prev_caller = fetch_reg(31, regs);

	/*
	 * Emulation code
	 */

	/* handle cas, ldstub, swap */
	if (dir == both) {
		unsigned short op3 = (insn >> 19) & 0x3f;
		mcd_emulate_ldst++;
		if (op3 == 0x3c || op3 == 0x3e) {
			/* cas/casx: requires special computation for effective address */
			unsigned int rs1 = (insn >> 14) & 0x1f;
			unsigned int rs2 = insn & 0x1f;
			unsigned int rd  = (insn >> 25) & 0x1f;
			int from_kernel = (regs->tstate & TSTATE_PRIV) != 0;
			unsigned long rs2_val, rd_val, *rd_addr;

			maybe_flush_windows(rs1, rs2, rd, from_kernel);
			addr = fetch_reg(rs1, regs);
			if (addr != fault_addr) {
				printk(KERN_ALERT "%s: cas/casx computed addr 0x%lx != fault addr 0x%lx, failing\n",
				       __func__, addr, fault_addr);
				ret = 0;
				goto err;
			}
			if (asi != 0x80) {
				printk(KERN_ALERT "%s: cas/casx with ASI 0x%x, failing\n", __func__, asi);
				ret = 0;
				goto err;
			}
			rs2_val = fetch_reg(rs2, regs);
			rd_val =  fetch_reg(rd,  regs);
			if (op3 == 0x3c)
				__asm__ __volatile__(
					"casa [%1] 0x80, %2, %0\n\t"
					: "+r" (rd_val)
					: "r" (addr), "r" (rs2_val));
			else
				__asm__ __volatile__(
					"casxa [%1] 0x80, %2, %0\n\t"
					: "+r" (rd_val)
					: "r" (addr), "r" (rs2_val));
			rd_addr = fetch_reg_addr(rd, regs);
			old_val = *rd_addr;
			*rd_addr = rd_val;

			mcd_rec.rs1_val = addr;
			mcd_rec.rs2_val = rs2_val;
			mcd_rec.rd_val = rd_val;
			mcd_rec.old_rd_val = old_val;
			mcd_rec.count = mcd_emulate_ldst;

			if (las_cases++ > 1000000) {
				static int stucked = 0;
				printk(KERN_ALERT "%s: STUCK for 1,000,000 instructions\n", __func__);
				(void)sprint_symbol(symbuf, regs->tpc);
				printk(KERN_ALERT "\t\t%lx\tcas%sa\t[%lx] #%x %%r%d, %%r%d\t= 0x%016lx\t(was: 0x%016lx)\t%ld+(%s)\n",
				       regs->tpc, (op3 == 0x3e) ? "x" : "",
				       addr, asi, rs2, rd, rd_val, old_val, now-timestamp, symbuf);
				/*
				 * we can't actually do this... (why not?)
				printk(KERN_ALERT "%s: *** DISABLING MCD ***\n", __func__);
				mcd_enable = 0;
				mcd_master_enable = 0;
				ret = 1;
				*/
				if (stucked++ > 100) {
					ret = 0;
					goto err;
				}
			}

			ret = 1;
			goto out_casa;
		}
		if (op3 == 0xd || op3 == 0x1d) {
			/* ldstub/ldstuba */
			printk(KERN_ALERT "%s: LDSTUB unhandled (op3=0x%x)\n", __func__, op3);
			ret = 0;
			goto err;
		}
		if (op3 == 0xf || op3 == 0x1f) {
			/* swap/swapa */
			printk(KERN_ALERT "%s: SWAP unhandled (op3=0x%x)\n", __func__, op3);
			ret = 0;
			goto err;
		}
		printk(KERN_ALERT "%s: unknown instruction 0x%x (op3=0x%x)\n", __func__, insn, op3);
		ret = 0;
		goto err;
	}

	if (las_cases > 100)
		printk(KERN_ALERT "\n%s: emulated %d consecutive CAS instructions\n", __func__, las_cases);

	/* emulate loads */
	if (dir == load) {
		if (addr != fault_addr) { /* should not happen */
			printk(KERN_ALERT "%s: computed load addr 0x%lx != fault addr 0x%lx, failing",
			       __func__, addr, fault_addr);
			ret = 0;
			goto err;
		}
		mcd_emulate_loads++;
		reg_addr = fetch_reg_addr(((insn>>25)&0x1f), regs);
		old_val = *reg_addr;

		switch (access_size) {
			case 8:
				reg_val = *(unsigned long *)addr;
				break;
			case 4:
				if (sign)
					reg_val = *(signed int *)addr;
				else
					reg_val = *(unsigned int *)addr;
				old_val = old_val & 0xffffffff;
				break;
			case 2:
				if (sign)
					reg_val = *(signed short *)addr;
				else
					reg_val = *(unsigned short *)addr;
				old_val = old_val & 0xffff;
				break;
			case 1:
				if (sign)
					reg_val = *(signed char *)addr;
				else
					reg_val = *(unsigned char *)addr;
				old_val = old_val & 0xff;
				break;
			default:
				ret = 0;	/* panic */
				goto err;
		}

		mcd_rec.rd_val = reg_val;
		mcd_rec.old_rd_val = old_val;
		mcd_rec.count = mcd_emulate_loads;

		advance(regs);
		*reg_addr = reg_val;
		ret = 1;
		goto out;
	}

	/* prepare to emulate store */
	reg_val = fetch_reg(((insn>>25)&0x1f), regs);
	if (addr != fault_addr) { /* should not happen */
		printk(KERN_ALERT "%s: computed store addr 0x%lx != fault addr 0x%lx, failing",
		       __func__, addr, fault_addr);
		ret = 0;
		goto err;
	}

	mcd_emulate_stores++;

	switch (access_size) {
	case 8:
		old_val = *(unsigned long *)addr;
		*(unsigned long *)addr = reg_val;
		reg_val = *(unsigned long *)addr;
		break;
	case 4:
		old_val = *(unsigned int *)addr;
		reg_val = reg_val & 0xffffffff;
		*(unsigned int *)addr = (unsigned int) reg_val;
		reg_val = *(unsigned int *)addr;
		break;
	case 2:
		old_val = *(unsigned short *)addr;
		reg_val = reg_val & 0xffff;
		*(unsigned short *)addr = (unsigned short) reg_val;
		reg_val = *(unsigned short *)addr;
		break;
	case 1:
		old_val = *(unsigned char *)addr;
		reg_val = reg_val & 0xff;
		*(unsigned char *)addr = (unsigned char) reg_val;
		reg_val = *(unsigned char *)addr;
		break;
	default:
		ret = 0; /*panic */
		goto err;
	}

	mcd_rec.rd_val = reg_val;
	mcd_rec.old_mem_val = old_val;
	mcd_rec.new_mem_val = reg_val;
	mcd_rec.count = mcd_emulate_stores;

	ret = 1;
	advance(regs);

out:
	las_cases = 0;
out_casa:
	/*
	 * The heart of the watchpoint watcher:
	 * If we hit a watchpoint, invoke the owner's callback function
	 */
	w = find_watchpoint(addr, access_size, MATCH_OVERLAP);
	if (w) {
		if (w->fn) {
			w->fn(&mcd_rec, &watchpoint_services, w->private);
		}
		else {
			char symbuf[512];
			(void) sprint_symbol(symbuf, w->owner);
			printk(KERN_ALERT "Struck WATCHPOINT owned by %s\n", symbuf);
			watchpoint_print_record(&mcd_rec);
			(void) sprint_symbol(symbuf, mcd_rec.prev_caller);
			printk(KERN_ALERT "Called by %s\n", symbuf);
			ret = 0; /* will panic */
		}
	}

err:
	nesting--;
	RDTICK(timestamp);
	return ret;
}

/*
 * Public entry point for adding a watchpoint
 */
int register_watchpoint(void *addr, int len, watchpoint_callback_t fn, void *private)
{
	watchpoint_t *w;
	int i;
	char owner[512], callback[512];

	if (mcd_not_present())
		return -ENOTSUPP;

	// look for an existing watchpoint that covers or overlaps this range
	w = find_watchpoint((unsigned long) addr, len, MATCH_OVERLAP);

	// return error if one exists
	if (w) {
		printk(KERN_ALERT "%s: [%p,%d] conflicts with existing [0x%lx,%d]\n", __func__,
		       addr, len, w->addr, w->len);
		return -EEXIST;
	}

	// check watch address for validity?

	// allocate new watchpoint record
	for (i=0; i<NUM_WATCHPOINTS; i++) {
		w = &watchpoints[i];
		if (w->addr == 0) {
			unsigned long caller;

			if (set_mcd_tag_range((unsigned long) addr, len, WATCHPOINT_MEMORY_TAG) != WATCHPOINT_MEMORY_TAG) {
				printk(KERN_ALERT "%s: set_mcd_tag_range failed\n", __func__);
				return -ENOSYS;
			}
			w->addr = (unsigned long) addr;
			w->len = len;
			w->fn = fn;
			w->private = private;
			__asm__ __volatile__("mov %%i7, %0\n\t" : "=r" (caller) ::);
			w->owner = caller;
			(void)sprint_symbol(owner, caller);
			(void)sprint_symbol(callback, (unsigned long)fn);
			printk(KERN_ALERT "%s: Watchpoint [%p,%d] set by %s, callback=%s\n",
			       __func__, addr, len, owner, callback);

			mcd_enable = TSTATE_MCDE;
			mcd_master_enable = 1;
			set_mcde_state(1);
			return 0;
		}
	}
	return -ENOMEM;
}
EXPORT_SYMBOL(register_watchpoint);

/*
 * Public entry point for removing a watchpoint
 */
int deregister_watchpoint(void *addr, int len)
{
	watchpoint_t *w = find_watchpoint((unsigned long)addr, len, MATCH_EXACT);
	char owner[512], remover[512];
	unsigned long caller;

	if (mcd_not_present())
		return -ENOTSUPP;

	if (w == NULL) {
		printk(KERN_ALERT "%s: Watchpoint [%p,%d] could not be found\n",
		       __func__, addr, len);
		return -EEXIST;
	}

	if (set_mcd_tag_range((unsigned long) w->addr, w->len, 0) != 0) {
		printk(KERN_ALERT "%s: failed to unset memory version for 0x%lx,%d\n",
		       __func__, w->addr, w->len);
		/* this is actually a really bad thing to happen, and we ought to panic */
	}

	w->addr = 0;
	w->len = 0;
	w->fn = 0;

	(void)sprint_symbol(owner, w->owner);
	__asm__ __volatile__("mov %%i7, %0\n\t" : "=r" (caller) ::);
	(void)sprint_symbol(remover, caller);
	printk(KERN_ALERT "%s: Watchpoint [%p,%d] installed by %s removed by %s\n",
	       __func__, addr, len, owner, remover);

	w->owner = 0;

	/* disable MCD entirely if there are no watchpoints active */
	if (find_watchpoint(0, 0xffffffff, MATCH_OVERLAP) == NULL) {
		mcd_enable = 0;
		mcd_master_enable = 0;
		disable_mcde();
	}
	return 0;
}
EXPORT_SYMBOL(deregister_watchpoint);

/* this is for test code to use */
char watchpoint_test_space[512];
EXPORT_SYMBOL(watchpoint_test_space);
