/*
 * SPARC64 Huge TLB page support.
 *
 * Copyright (C) 2002, 2003, 2006 David S. Miller (davem@davemloft.net)
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/hugetlb.h>
#include <linux/pagemap.h>
#include <linux/sysctl.h>
#include <asm/mman.h>
#include <asm/pgalloc.h>
#include <asm/tlb.h>
#include <asm/tlbflush.h>
#include <asm/cacheflush.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>

static unsigned int xl_hugepages_debug;
static int __init xl_hugepages_debug_setup(char *str)
{
	xl_hugepages_debug = 1U;
	return 1;
}
__setup("xl_hugepages_debug=", xl_hugepages_debug_setup);

/* Slightly simplified from the non-hugepage variant because by
 * definition we don't have to worry about any page coloring stuff
 */

static unsigned long hugetlb_get_unmapped_area_bottomup(struct file *file,
							unsigned long addr,
							unsigned long len,
							unsigned long pgoff,
							unsigned long flags)
{
	struct hstate *h = hstate_file(file);
	struct mm_struct *mm = current->mm;
	struct vm_area_struct * vma;
	unsigned long task_size = TASK_SIZE;
	unsigned long start_addr;

	if (test_thread_flag(TIF_32BIT))
		task_size = STACK_TOP32;
	if (unlikely(len >= VA_EXCLUDE_START))
		return -ENOMEM;

	if (len > mm->cached_hole_size) {
	        start_addr = addr = mm->free_area_cache;
	} else {
	        start_addr = addr = TASK_UNMAPPED_BASE;
	        mm->cached_hole_size = 0;
	}

	task_size -= len;

full_search:
	addr = ALIGN(addr, huge_page_size(h));

	for (vma = find_vma(mm, addr); ; vma = vma->vm_next) {
		/* At this point:  (!vma || addr < vma->vm_end). */
		if (addr < VA_EXCLUDE_START &&
		    (addr + len) >= VA_EXCLUDE_START) {
			addr = VA_EXCLUDE_END;
			vma = find_vma(mm, VA_EXCLUDE_END);
		}
		if (unlikely(task_size < addr)) {
			if (start_addr != TASK_UNMAPPED_BASE) {
				start_addr = addr = TASK_UNMAPPED_BASE;
				mm->cached_hole_size = 0;
				goto full_search;
			}
			return -ENOMEM;
		}
		if (likely(!vma || addr + len <= vma->vm_start)) {
			/*
			 * Remember the place where we stopped the search:
			 */
			mm->free_area_cache = addr + len;
			return addr;
		}
		if (addr + mm->cached_hole_size < vma->vm_start)
		        mm->cached_hole_size = vma->vm_start - addr;

		addr = ALIGN(vma->vm_end, huge_page_size(h));
	}
}

static unsigned long
hugetlb_get_unmapped_area_topdown(struct file *file, const unsigned long addr0,
				  const unsigned long len,
				  const unsigned long pgoff,
				  const unsigned long flags)
{
	struct hstate *h = hstate_file(file);
	struct vm_area_struct *vma;
	struct mm_struct *mm = current->mm;
	unsigned long addr = addr0;

	/* This should only ever run for 32-bit processes.  */
	BUG_ON(!test_thread_flag(TIF_32BIT));

	/* check if free_area_cache is useful for us */
	if (len <= mm->cached_hole_size) {
 	        mm->cached_hole_size = 0;
 		mm->free_area_cache = mm->mmap_base;
 	}

	/* either no address requested or can't fit in requested address hole */
	addr = mm->free_area_cache & huge_page_mask(h);

	/* make sure it can fit in the remaining address space */
	if (likely(addr > len)) {
		vma = find_vma(mm, addr-len);
		if (!vma || addr <= vma->vm_start) {
			/* remember the address as a hint for next time */
			return (mm->free_area_cache = addr-len);
		}
	}

	if (unlikely(mm->mmap_base < len))
		goto bottomup;

	addr = (mm->mmap_base-len) & huge_page_mask(h);

	do {
		/*
		 * Lookup failure means no vma is above this address,
		 * else if new region fits below vma->vm_start,
		 * return with success:
		 */
		vma = find_vma(mm, addr);
		if (likely(!vma || addr+len <= vma->vm_start)) {
			/* remember the address as a hint for next time */
			return (mm->free_area_cache = addr);
		}

 		/* remember the largest hole we saw so far */
 		if (addr + mm->cached_hole_size < vma->vm_start)
 		        mm->cached_hole_size = vma->vm_start - addr;

		/* try just below the current vma->vm_start */
		addr = (vma->vm_start-len) & huge_page_mask(h);
	} while (likely(len < vma->vm_start));

bottomup:
	/*
	 * A failed mmap() very likely causes application failure,
	 * so fall back to the bottom-up function here. This scenario
	 * can happen with large stack limits and large mmap()
	 * allocations.
	 */
	mm->cached_hole_size = ~0UL;
  	mm->free_area_cache = TASK_UNMAPPED_BASE;
	addr = arch_get_unmapped_area(file, addr0, len, pgoff, flags);
	/*
	 * Restore the topdown base:
	 */
	mm->free_area_cache = mm->mmap_base;
	mm->cached_hole_size = ~0UL;

	return addr;
}

unsigned long hugetlb_get_unmapped_area(struct file *file, unsigned long addr,
		unsigned long len, unsigned long pgoff, unsigned long flags)
{
	struct hstate *h = hstate_file(file);
	unsigned long task_size = TASK_SIZE;
	struct mm_struct *mm = current->mm;
	struct vm_area_struct *vma;

	if (test_thread_flag(TIF_32BIT))
		task_size = STACK_TOP32;

	if (len & ~huge_page_mask(h))
		return -EINVAL;
	if (len > task_size)
		return -ENOMEM;

	if (flags & MAP_FIXED) {
		if (prepare_hugepage_range(file, addr, len))
			return -EINVAL;
		return addr;
	}

	if (addr) {
		addr = ALIGN(addr, huge_page_size(h));
		vma = find_vma(mm, addr);
		if (task_size - len >= addr &&
		    (!vma || addr + len <= vma->vm_start))
			return addr;
	}
	if (mm->get_unmapped_area == arch_get_unmapped_area)
		return hugetlb_get_unmapped_area_bottomup(file, addr, len,
				pgoff, flags);
	else
		return hugetlb_get_unmapped_area_topdown(file, addr, len,
				pgoff, flags);
}

/* Since the hugepage could cover more than one pmd entry and more
 * than one pgd entry we must cover all possible conditions.
 */
static pmd_t *huge_pmd_alloc(struct mm_struct *mm, unsigned long addr)
{
	pgd_t *pgd = pgd_offset(mm, addr);
	pmd_t *pmd = NULL;
	pud_t *pud;

	if (pgd_none(*pgd)) {
		pud_t *pud = pud_alloc(mm, pgd, addr);

		if (pud == NULL)
			goto out;
	}
	pud = pud_offset(pgd, addr);
	if (pud_none(*pud)) {
		pmd = pmd_alloc(mm, pud, addr);

		if (pmd == NULL)
			goto out;
	}
	pmd = pmd_offset(pud, addr);
out:
	return pmd;
}

/*
 * Note, should we fail leave behind the mm state
 * which will be cleaned up on exit.
 */
pte_t *huge_pte_alloc(struct mm_struct *mm, unsigned long addr,
			unsigned long size)
{
	unsigned long start = addr & ~(size - 1);
	unsigned long end = start + size;
	pte_t *rpte = NULL;

	if (xl_hugepages_debug)
		printk(KERN_INFO
			"huge_pte_alloc: addr=0x%lx size=0x%lx\n", addr, size);

	/* Our caller operates on start's pte which is rpte should we succeed.*/
	for (addr = start; addr < end; addr = addr + PMD_SIZE) {
		pte_t *pte;
		pmd_t *pmd = huge_pmd_alloc(mm, addr);
		if (!pmd)
			goto fail;
		if (size != (1UL << 34)) {
			rpte = (pte_t *)pmd;
			break;
		}

		pte = pte_alloc_map(mm, NULL, pmd, addr);

		if (!pte)
			goto fail;
		else if (!rpte)
			rpte = pte;
	}

	if (xl_hugepages_debug)
		printk(KERN_INFO
			"huge_pte_alloc: rpte=0x%lx\n", (unsigned long) rpte);
	return rpte;
fail:
	return NULL;
}

int huge_pmd_unshare(struct mm_struct *mm, unsigned long *addr, pte_t *ptep)
{
	return 0;
}

/*
 * This function possibly needs to be moved. It will be different
 * for sun4u and even possibly for sun4v future cores. Though we have
 * no plans to support sun4u at this point.
 */
static unsigned int sun4v_tte_to_shift(pte_t entry)
{
	unsigned long hugepage_tte = pte_val(entry) & _PAGE_SZALL_4V;
	unsigned int hugepage_shift;

	switch (hugepage_tte) {
	case _PAGE_SZ16GB_4V:
		hugepage_shift = 34U;
		break;
	case _PAGE_SZ2GB_4V:
		hugepage_shift = 31U;
		break;
	case _PAGE_SZ256MB_4V:
		hugepage_shift = 28U;
		break;
	case _PAGE_SZ4MB_4V:
		hugepage_shift = 22U;
		break;
	default:
		WARN_ONCE(1, "hugepage_shift: hugepage_tte=0x%lx\n",
			hugepage_tte);
		hugepage_shift = PAGE_SHIFT;
		break;
	}
	return hugepage_shift;
}

static unsigned int tte_to_shift(pte_t entry)
{
	unsigned int hugepage_shift;

	if (tlb_type == hypervisor)
		hugepage_shift = sun4v_tte_to_shift(entry);
	else
		hugepage_shift = REAL_HPAGE_SHIFT;

	return hugepage_shift;
}

static unsigned long tte_to_hugepage_size(pte_t pte)
{
	unsigned long hugepage_size = 1UL << tte_to_shift(pte);

	if (hugepage_size == REAL_HPAGE_SIZE)
		hugepage_size = HPAGE_SIZE;
	return hugepage_size;
}

static unsigned long tte_to_hugepage_mask(pte_t pte)
{
	unsigned int hugepage_shift = tte_to_shift(pte);
	unsigned long hugepage_mask;

	if (hugepage_shift == REAL_HPAGE_SHIFT)
		hugepage_shift = HPAGE_SHIFT;

	hugepage_mask = ~((1UL << hugepage_shift) - 1);

	return hugepage_mask;
}

/*
 * This should also be moved and a noop for sun4u.
 * Only include hugepage sizes we plan to support.
 */
static pte_t hugepage_shift_to_tte(pte_t entry, unsigned int hugepage_shift)
{
	unsigned long sun4v_hugepage_size = _PAGE_SZ4MB_4V;

	pte_val(entry) = pte_val(entry) & ~_PAGE_SZALL_4V;

	switch (hugepage_shift) {
	/* 16Gb */
	case 34U:
		sun4v_hugepage_size = _PAGE_SZ16GB_4V;
		break;
	/* 2Gb */
	case 31U:
		sun4v_hugepage_size = _PAGE_SZ2GB_4V;
		pte_val(entry) |= _PAGE_PMD_HUGE;
		break;
	/* 8M */
	case 23U:
		pte_val(entry) |= _PAGE_PMD_HUGE;
		break;
	default:
		WARN_ONCE(hugepage_shift,
			"hugepage_shift_to_tte: unsupported "
			"hugepage_shift=%u.\n", hugepage_shift);
	}

	pte_val(entry) = pte_val(entry) | sun4v_hugepage_size;
	return entry;
}

pte_t arch_make_huge_pte(pte_t entry, struct vm_area_struct *vma,
				struct page *page, int writeable)
{
	unsigned int hugepage_shift = huge_page_shift(hstate_vma(vma));

	entry = hugepage_shift_to_tte(entry, hugepage_shift);

#ifdef CONFIG_SPARC_ADI
	/* If this vma has ADI enabled on it, turn on TTE.mcd
	 */
	if (vma->vm_flags & VM_SPARC_ADI)
		return pte_mkmcd(entry);
	else
		return pte_mknotmcd(entry);
#else
	return entry;
#endif
}

struct set_at_data {
	pte_t	orig;	/* Original pte value before update or sentinel.*/
};

static void huge_pte_at_flush_update(struct mm_struct *mm, unsigned long addr,
				pte_t *pte, pte_t orig,
				struct set_at_data *sa_data)
{
	if (pte_val(orig) & _PAGE_VALID) {
		if (!(pte_val(sa_data->orig) & _PAGE_VALID)) {
			sa_data->orig = orig;
			tlb_batch_add(mm, addr, pte, orig, false);
		}
	}
}

static void form_sentinel(struct set_at_data *sa_data, pte_t entry, pte_t *pte)
{
	pte_val(sa_data->orig) = _PAGE_VALID | _PAGE_SPECIAL_4V |
		_PAGE_PMD_HUGE | (pte_val(entry) & _PAGE_SZALL_4V);
	if (xl_hugepages_debug)
		pr_info("form_sentinel: entry=0x%lx, pte=0x%lx, "
			"base pte PA=0x%lx\n", pte_val(entry),
			(unsigned long)pte, __pa(pte));
	BUG_ON(__pa(pte) & _PAGE_SZALL_4V);
	pte_val(sa_data->orig) = pte_val(sa_data->orig) | __pa(pte);
}

static bool huge_pte_at_handle_sentinel(struct set_at_data *sa_data,
				pte_t *pte, pte_t orig, pte_t entry)
{
	bool rc = true;

	/*
	 * Should the original pte be marked valid then
	 * only update the sentinel.
	 */
	if (pte_val(orig) & _PAGE_VALID) {
		if ((pte_val(orig) & _PAGE_SPECIAL_4V) == 0UL)
			*pte = entry;
		rc = false;
	} else if (pte_val(sa_data->orig) & _PAGE_VALID) {
		*pte = sa_data->orig;
	} else {
		form_sentinel(sa_data, entry, pte);
		*pte = entry;
	}

	return rc;
}

static bool __set_huge_pte_at(struct mm_struct *mm, unsigned long addr,
			pte_t *pte, pte_t entry,
			struct set_at_data *sa_data,
			unsigned int hugepage_shift)
{
	bool rc = true;

	if (hugepage_shift != REAL_HPAGE_SHIFT) {
		pte_t orig = *pte;

		rc = huge_pte_at_handle_sentinel(sa_data, pte, orig, entry);
		huge_pte_at_flush_update(mm, addr, pte, orig, sa_data);
	} else {
		*pte = entry;
	}

	return rc;
}

static void __clear_huge_pte_at(struct mm_struct *mm, unsigned long addr,
			pte_t *pte, struct set_at_data *sa_data,
			unsigned int hugepage_shift)
{
	if (hugepage_shift != REAL_HPAGE_SHIFT) {
		pte_t orig = *pte;

		*pte = __pte(0UL);
		huge_pte_at_flush_update(mm, addr, pte, orig, sa_data);
	} else {
		*pte = __pte(0UL);
	}
}

static bool set_huge_pte_range_at(struct mm_struct *mm, pmd_t *pmd,
			unsigned long addr, pte_t *pentry,
			struct set_at_data *sa_data, bool set_at)
{
	pte_t *pte = pte_offset_map(pmd, addr);
	pte_t *lpte = pte + PTRS_PER_PTE;
	pte_t entry = *pentry;
	bool rc = true;
	unsigned int hugepage_shift;

	if (set_at)
		hugepage_shift = tte_to_shift(entry);
	else
		hugepage_shift = tte_to_shift(*pte);

	for (; pte < lpte; pte++, addr = addr + PAGE_SIZE) {
		if (set_at) {
			rc = __set_huge_pte_at(mm, addr, pte, entry, sa_data,
					hugepage_shift);
			if (!rc)
				break;
			pte_val(entry) = pte_val(entry) + PAGE_SIZE;
		} else
			__clear_huge_pte_at(mm, addr, pte, sa_data,
					hugepage_shift);
	}
	if (set_at)
		*pentry = entry;

	return rc;
}

static bool __set_huge_pmd_at(struct mm_struct *mm, pmd_t *pmd,
		unsigned long addr, unsigned long end, pte_t *pentry,
		struct set_at_data *sa_data, bool set_at)
{
	bool rc;
	pte_t orig;
	pte_t entry;
	unsigned long next;
	unsigned long hugepage_shift;

	rc = true;
	orig = *(pte_t *)pmd;
	entry = *pentry;

	if (set_at) {
		hugepage_shift = tte_to_shift(entry);
		if (hugepage_shift == REAL_HPAGE_SHIFT) {
			*pmd = __pmd(pte_val(entry));
		} else {
			do {
				next = pmd_addr_end(addr, end);
				rc = __set_huge_pte_at(mm, addr, (pte_t *)pmd,
					entry, sa_data,	hugepage_shift);
				if (!rc)
					break;
			} while (pmd++, addr = next, addr != end);
		}
		*pentry = entry;
	} else {
		hugepage_shift = tte_to_shift(orig);
		if (hugepage_shift == REAL_HPAGE_SHIFT) {
			*pmd = __pmd(0);
		} else {
			do {
				next = pmd_addr_end(addr, end);
				__clear_huge_pte_at(mm, addr, (pte_t *)pmd,
					sa_data, hugepage_shift);
			} while (pmd++, addr = next, addr != end);
		}
	}

	if (hugepage_shift == REAL_HPAGE_SHIFT) {
		/* Issue TLB flush at REAL_HPAGE_SIZE boundaries */
		maybe_tlb_batch_add(mm, addr, (pte_t *)pmd, orig, 0);
		maybe_tlb_batch_add(mm, addr + REAL_HPAGE_SIZE,
					(pte_t *)pmd, orig, 0);
	}

	return rc;
}

static bool set_huge_pmd_at(struct mm_struct *mm, pud_t *pud,
		unsigned long addr, unsigned long end, pte_t *pentry,
		struct set_at_data *sa_data, bool set_at)
{
	pmd_t *pmd = pmd_offset(pud, addr);
	unsigned long next;
	bool rc;
	pmd_t orig;
	pmd_t entry;
	unsigned int is_huge_pmd;

	orig = *pmd;
	entry = __pmd(pte_val(*pentry));

	if (set_at)
		is_huge_pmd = is_hugetlb_pmd(entry);
	else
		is_huge_pmd = is_hugetlb_pmd(orig);

	if (is_huge_pmd)
		return __set_huge_pmd_at(mm, pmd, addr, end, pentry, sa_data,
					set_at);

	do {
		next = pmd_addr_end(addr, end);
		rc = set_huge_pte_range_at(mm, pmd, addr, pentry,
				sa_data, set_at);
	} while (pmd++, addr = next, ((addr != end) && rc));
	return rc;
}

static bool set_huge_pud_at(struct mm_struct *mm, pgd_t *pgd,
			unsigned long addr, unsigned long end, pte_t *pentry,
			struct set_at_data *sa_data, bool set_at)
{
	pud_t *pud = pud_offset(pgd, addr);
	unsigned long next;
	bool rc;

	do {
		next = pud_addr_end(addr, end);
		rc = set_huge_pmd_at(mm, pud, addr, next, pentry,
				sa_data, set_at);
	} while (pud++, addr = next, ((addr != end) && rc));
	return rc;
}

static void xl_hugepages_entry_info(char ch, unsigned long hugepage_size,
		unsigned long hugepage_mask, pte_t entry, unsigned long addr,
		unsigned long start, unsigned long end)
{
	if (!xl_hugepages_debug)
		goto out;

	printk(KERN_INFO
		"%c hugepage_size=0x%lx hugepage_mask=0x%lx entry=0x%lx\n",
		ch, hugepage_size, hugepage_mask, pte_val(entry));
	printk(KERN_INFO "%c addr=0x%lx start=0x%lx end=0x%lx\n",
		ch, addr, start, end);
out:
	return;
}

/*
 * entry must be the first pte of the hugepage. Otherwise entry
 * must be adjusted before we enter the loop for set_pte_at and
 * aligned physically to match the hugepage_size. This is equally
 * true of other locations where HUGETLB_PAGE_ORDER is used within
 * this module for mainline as of 7/20/2014.
 */
void set_huge_pte_at(struct mm_struct *mm, unsigned long addr,
		     pte_t *ptep, pte_t entry)
{
	struct set_at_data sa_data = {.orig = __pte(0UL)};
	unsigned long hugepage_size = tte_to_hugepage_size(entry);
	unsigned long hugepage_mask = tte_to_hugepage_mask(entry);
	unsigned long start = addr & hugepage_mask;
	unsigned long end = start + hugepage_size;
	pgd_t *pgd = pgd_offset(mm, start);
	unsigned long next;
	bool rc;

	if (!pte_present(*ptep) && pte_present(entry)) {
		unsigned int pte_count_idx =
			hugepage_size_to_pte_count_idx(hugepage_size);

		mm->context.huge_pte_count[pte_count_idx]++;
	}

	xl_hugepages_entry_info('S', hugepage_size, hugepage_mask, entry,
		addr, start, end);

	WARN_ONCE(((pte_pfn(entry) << PAGE_SHIFT) & ~hugepage_mask),
		"set_huge_pte_at: ptep not aligned, pte=0x%lx to"
		" hugepage_mask=0x%lx.\n",
		pte_val(entry), hugepage_mask);

	do {
		next = pgd_addr_end(start, end);
		rc = set_huge_pud_at(mm, pgd, start, next, &entry,
						&sa_data, true);
	} while(pgd++, start = next, ((start != end) && rc));
}

static pte_t huge_ptep_get_and_clear_all(struct mm_struct *mm,
					 unsigned long addr, pte_t entry)
{
	struct set_at_data sa_data = {.orig = __pte(0UL)};
	unsigned long hugepage_size = tte_to_hugepage_size(entry);
	unsigned long hugepage_mask = tte_to_hugepage_mask(entry);
	unsigned long start = addr & hugepage_mask;
	unsigned long end = start + hugepage_size;
	pgd_t *pgd = pgd_offset(mm, start);
	unsigned long next;
	bool rc;

	if (pte_present(entry)) {
		unsigned int pte_count_idx =
			hugepage_size_to_pte_count_idx(hugepage_size);

		mm->context.huge_pte_count[pte_count_idx]--;
	}

	xl_hugepages_entry_info('C', hugepage_size, hugepage_mask, entry,
		addr, start, end);

	WARN_ONCE(((pte_pfn(entry) << PAGE_SHIFT) & ~hugepage_mask),
		"huge_ptep_get_and_clear: ptep not aligned, pte=0x%lx, to"
		" hugepage_mask=0x%lx.\n",
		pte_val(entry), hugepage_mask);

	do {
		next = pgd_addr_end(start, end);
		rc = set_huge_pud_at(mm, pgd, start, next, &entry,
						&sa_data, false);
	} while(pgd++, start = next, ((start != end) && rc));

	return entry;
}

pte_t huge_ptep_get_and_clear(struct mm_struct *mm, unsigned long addr,
				pte_t *ptep)
{
	pte_t entry = *ptep;

	if (!huge_pte_none(entry))
		entry = huge_ptep_get_and_clear_all(mm, addr, entry);

	return entry;
}

unsigned long xl_hugepage_mask = HPAGE_MASK;

pte_t *huge_pte_offset(struct mm_struct *mm, unsigned long addr)
{
	pte_t *pte = NULL;
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;

	addr = addr & xl_hugepage_mask;

	pgd = pgd_offset(mm, addr);
	if (!pgd_none(*pgd)) {
		pud = pud_offset(pgd, addr);
		if (!pud_none(*pud)) {
			pmd = pmd_offset(pud, addr);
			if (xl_hugepage_shift != 34U)	/* 8M or 2G page */
				pte = (pte_t *)pmd;
			else if (!pmd_none(*pmd))	/* 16G page */
				pte = pte_offset_map(pmd, addr);
		}
	}


	if (pte && (pte_val(*pte) & _PAGE_VALID)) {
		unsigned long hugepage_mask = tte_to_hugepage_mask(*pte);

		WARN_ONCE(((pte_pfn(*pte) << PAGE_SHIFT) & ~hugepage_mask),
			"huge_pte_offset: ptep not aligned, pte=0x%lx to"
			" hugepage_mask=0x%lx.\n",
			pte_val(*pte), hugepage_mask);

		BUG_ON(pte_val(*pte) & _PAGE_SPECIAL_4V);
	}

	return pte;
}

struct page *follow_huge_addr(struct mm_struct *mm,
			      unsigned long address, int write)
{
	return ERR_PTR(-EINVAL);
}

int pmd_huge(pmd_t pmd)
{
	return !pmd_none(pmd) &&
		(pmd_val(pmd) & (_PAGE_VALID|_PAGE_PMD_HUGE)) != _PAGE_VALID;
}

int pud_huge(pud_t pud)
{
	return 0;
}

struct page *follow_huge_pmd(struct mm_struct *mm, unsigned long address,
			     pmd_t *pmd, int write)
{
	return NULL;
}


static void hugetlb_free_pte_range(struct mmu_gather *tlb, pmd_t *pmd,
			   unsigned long addr)
{
	pgtable_t token = pmd_pgtable(*pmd);
	pmd_clear(pmd);
	pte_free_tlb(tlb, token, addr);
	tlb->mm->nr_ptes--;
}

static void hugetlb_free_pmd_range(struct mmu_gather *tlb, pud_t *pud,
				   unsigned long addr, unsigned long end,
				   unsigned long floor, unsigned long ceiling)
{
	pmd_t *pmd;
	unsigned long next;
	unsigned long start;

	start = addr;
	pmd = pmd_offset(pud, addr);
	do {
		next = pmd_addr_end(addr, end);
		if (pmd_none(*pmd))
			continue;
		if (is_hugetlb_pmd(*pmd))
			pmd_clear(pmd);
		else
			hugetlb_free_pte_range(tlb, pmd, addr);
	} while (pmd++, addr = next, addr != end);

	start &= PUD_MASK;
	if (start < floor)
		return;
	if (ceiling) {
		ceiling &= PUD_MASK;
		if (!ceiling)
			return;
	}
	if (end - 1 > ceiling - 1)
		return;

	pmd = pmd_offset(pud, start);
	pud_clear(pud);
	pmd_free_tlb(tlb, pmd, start);
}

static void hugetlb_free_pud_range(struct mmu_gather *tlb, pgd_t *pgd,
				   unsigned long addr, unsigned long end,
				   unsigned long floor, unsigned long ceiling)
{
	pud_t *pud;
	unsigned long next;
	unsigned long start;

	start = addr;
	pud = pud_offset(pgd, addr);
	do {
		next = pud_addr_end(addr, end);
		if (pud_none_or_clear_bad(pud))
			continue;
		hugetlb_free_pmd_range(tlb, pud, addr, next, floor,
				       ceiling);
	} while (pud++, addr = next, addr != end);

	start &= PGDIR_MASK;
	if (start < floor)
		return;
	if (ceiling) {
		ceiling &= PGDIR_MASK;
		if (!ceiling)
			return;
	}
	if (end - 1 > ceiling - 1)
		return;

	pud = pud_offset(pgd, start);
	pgd_clear(pgd);
	pud_free_tlb(tlb, pud, start);
}

void hugetlb_free_pgd_range(struct mmu_gather *tlb,
			    unsigned long addr, unsigned long end,
			    unsigned long floor, unsigned long ceiling)
{
	pgd_t *pgd;
	unsigned long next;

	pgd = pgd_offset(tlb->mm, addr);
	do {
		next = pgd_addr_end(addr, end);
		if (pgd_none_or_clear_bad(pgd))
			continue;
		hugetlb_free_pud_range(tlb, pgd, addr, next, floor, ceiling);
	} while (pgd++, addr = next, addr != end);
}
