
/*
 * Oracle M7 Database Analytics Accelerator (DAX) Driver
 */

/*
 * Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
 * GPL/UPL notice goes here (how)
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <linux/uaccess.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/mm.h>
#include <linux/kallsyms.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/bug.h>
#include <linux/hugetlb.h>
#include <linux/spinlock.h>
#include <linux/nodemask.h>
#include "dax.h"
#include "dax_impl.h"
#include "ccb.h"
#include <asm/hypervisor.h>
#include <asm/pcr.h>
#include <asm/pgtable.h>
#include <asm/mdesc.h>
#include <asm/atomic.h>

#define ASSERT(X) BUG_ON(!(X))

#define	DAX_DBG_FLG_BASIC	0x01
#define	DAX_DBG_FLG_DRV		0x02
#define	DAX_DBG_FLG_MAP		0x04
#define DAX_DBG_FLG_LIST        0x08
#define DAX_DBG_FLG_PERF        0x10
#define	DAX_DBG_FLG_ALL		0xff

#define print_err(fmt, ...) \
	printk(KERN_INFO fmt "\n", ##__VA_ARGS__)
#define	DAX_DBG			if (dax_debug & DAX_DBG_FLG_BASIC) print_err
#define	DAX_DBG_DRV		if (dax_debug & DAX_DBG_FLG_DRV)   print_err
#define	DAX_DBG_MAP		if (dax_debug & DAX_DBG_FLG_MAP)   print_err
#define	DAX_DBG_LIST		if (dax_debug & DAX_DBG_FLG_LIST)  print_err
#define	DAX_DBG_PERF		if (dax_debug & DAX_DBG_FLG_PERF)  print_err

static int dax_debug = 0;


#define	DAX_NAME		"dax"

#define	DAX_CCB_BUF_SZ		8*1024
#define	DAX_CCB_BUF_NELEMS	(DAX_CCB_BUF_SZ / sizeof (ccb_t))

#define	DAX_CA_BUF_SZ		(DAX_CCB_BUF_NELEMS * \
				 sizeof (ccb_completion_area_t))

#define	DAX_MMAP_SZ		DAX_CA_BUF_SZ
#define	DAX_MMAP_OFF		(off_t)0x0

#define	DWORDS_PER_CCB		8

#define	CCB_HDR(ccb)		((ccb_hdr_t *)(ccb))
#define	IS_LONG_CCB(ccb)	((CCB_HDR(ccb))->sync_flags & CCB_SYNC_LONGCCB)

#define	DAX_CCB_WAIT_USEC		100
#define	DAX_CCB_WAIT_RETRIES_MAX	10000

static long dax_version = DAX_DRIVER_VERSION;
static uint32_t dax_hv_ccb_submit_maxlen;
int dax_ccb_wait_usec = DAX_CCB_WAIT_USEC;
int dax_ccb_wait_retries_max = DAX_CCB_WAIT_RETRIES_MAX;
int m7_bug = 0;
atomic_t dax_alloc_counter = ATOMIC_INIT(0);
atomic_t dax_requested_mem = ATOMIC_INIT(0);
atomic_t dax_actual_mem = ATOMIC_INIT(0);
int dax_peak_waste = 0;

/*
 * CCB exec state
 */
typedef struct ccb_exec_ctl {
	uint64_t 	nomap_va;
	uint32_t        nretries;
	uint32_t	hwq_type;	/* msg/query */
	uint32_t	submitted_ccb_buf_len;
	uint64_t	submitted_ccb_buf_offset;
	uint32_t        ccb_exec_status;
} ccb_exec_ctl_t;

/* driver public entry points */
static int dax_open(struct inode *i, struct file *f);
static int dax_close(struct inode *i, struct file *f);
static long dax_ioctl(struct file *f, unsigned int cmd, unsigned long arg);
static int dax_devmap(struct file *filp, struct vm_area_struct *vma);
static ssize_t dax_read(struct file *, char __user *, size_t, loff_t *);

/* internal */
static dax_ctx_t *dax_ctx_alloc(void);
static int dax_ioctl_ccb_thr_init(void *, struct file *);
static int dax_ioctl_ccb_thr_fini(struct file *f);
static int dax_ioctl_ccb_exec(void *, struct file *);
static int dax_ioctl_ca_dequeue(void *, struct file *f);
static long perfcount_ioctl(struct file *f, unsigned int cmd, unsigned long arg);
static int dax_validate_ca_dequeue_args(dax_ctx_t *,
    dax_ca_dequeue_arg_t *);
static int dax_ccb_hv_submit(dax_ctx_t *, ccb_t *, size_t,
    ccb_exec_ctl_t *);
static int dax_validate_ccb_exec_args(dax_ccb_exec_arg_t *);
static int dax_validate_ccb(ccb_t *);
static ccb_t *dax_ccb_buffer_reserve(dax_ctx_t *, size_t, size_t *);
static void dax_ccb_buffer_commit(dax_ctx_t *, size_t);
static size_t dax_ccb_buffer_decommit(dax_ctx_t *, size_t);
static ccb_t *dax_ccb_buffer_get_contig_block(dax_ctx_t *, size_t *);
static void dax_map_segment(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len);
static void dax_unlock_pages(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len);
static int dax_lock_pages(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len);
static int dax_preprocess_usr_ccbs(dax_ctx_t *, ccb_t *, size_t);
static ccb_t *dax_ccb_buffer_get_contig_block(dax_ctx_t *, size_t *);
static void dax_ctx_fini(dax_ctx_t *);
static void dax_ctx_flush_decommit_ccbs(dax_ctx_t *);
static void dax_ctx_flush_ccbs(dax_ctx_t *);
static void dax_ccb_flush_contig(dax_ctx_t *dax, int, int);
static void dax_ccb_wait(dax_ctx_t *, int);
static void dax_state_destroy(struct file *f);
static void dax_prt_ccbs(ccb_t *, uint64_t);
#if 0
static int dax_vma_in_use(dax_vma_t *dv);
#endif
static void dax_vma_drain(dax_vma_t *dv);

static struct file_operations dax_fops = {
	.owner =    THIS_MODULE,
	.open =     dax_open,
	.release =  dax_close,
	.mmap =     dax_devmap,
	.read =     dax_read,
	.unlocked_ioctl = dax_ioctl
};

LIST_HEAD(dax_mm_list);
DEFINE_SPINLOCK(dax_mm_lock);

static int hv_get_hwqueue_size(unsigned long *qsize)
{
	long dummy;

	/* ccb = NULL, length = 0, Q type = query, VQ token = 0 */
	return sun4v_dax_ccb_submit(0, 0, HV_DAX_QUERY_CMD, 0, qsize, &dummy);
}

static dev_t first;
static struct cdev c_dev;
static struct class *cl;
static int force = 0;
module_param(force, int, 0644);
MODULE_PARM_DESC(force, "Forces module loading if no device present");
module_param(dax_debug, int, 0644);
MODULE_PARM_DESC(dax_debug, "Debug flags");

atomic_t has_erratum_6008  = ATOMIC_INIT(0);
atomic_t response_count = ATOMIC_INIT(0);

void dax_has_erratum_6008_client(void *info)
{
	extern int dax_has_erratum_6008_one_node(void);
	int cpu = smp_processor_id();
	int node = cpu_to_node(cpu);

	if (dax_has_erratum_6008_one_node()) {
		printk(KERN_INFO "DAX on cpu %d node %d has erratum 6008\n",
		       cpu, node);
		atomic_set(&has_erratum_6008, 1);
	} else {
		printk(KERN_INFO "DAX on cpu %d node %d has no erratum 6008\n",
		       cpu, node);
	}
	atomic_inc(&response_count);
}

int dax_has_erratum_6008_numa(void)
{
	unsigned int node;
	int cnt = 10000;
	int nr_nodes = 0;
	cpumask_t numa_cpu_mask;

	cpumask_clear(&numa_cpu_mask);
	atomic_set(&has_erratum_6008, 0);
	atomic_set(&response_count, 0);

	/*
	 * For M7 platforms with multi socket, processors on each socket may be
	 * of different version, thus different DAX version. So it is
	 * necessary to detect the erratum on all the DAXs in the
	 * platform. Select first cpu from each numa node and run the
	 * erratum detection code on those cpus. This makes sure
	 * that the detection code runs on all the DAXs in the platform.
	 */
	for_each_node_with_cpus(node) {
		int dst_cpu = first_cpu(numa_cpumask_lookup_table[node]);
		cpumask_set_cpu(dst_cpu, &numa_cpu_mask);
		nr_nodes++;
	}

	smp_call_function_many(&numa_cpu_mask,
			       dax_has_erratum_6008_client, NULL, 1);
	while ((atomic_read(&response_count) != nr_nodes) && --cnt)
		udelay(100);

	if (cnt == 0) {
		printk(KERN_WARNING "Could not synchronize DAX erratum detector\n");
		return 1;
	}

	return atomic_read(&has_erratum_6008);
}

static int __init
dax_attach(void)
{
	unsigned long stuff = 0, max_ccbs;
	int ret;
	struct mdesc_handle *hp = mdesc_grab();

	if (hp == NULL)
		printk(KERN_INFO "%s: Unable to grab mdesc\n", __func__);
	else {
		u64 pn;
		mdesc_for_each_node_by_name(hp, pn, "virtual-device") {
			int len;
			char *prop;
			prop = (char *) mdesc_get_property(hp, pn, "name", &len);
			if (prop == NULL)
				continue;
			if (strcmp(prop, "dax"))
				continue;
			printk(KERN_INFO "%s:   Found node 0x%Lx = %s\n", __func__, pn, prop);

			prop = (char *) mdesc_get_property(hp, pn, "compatible", &len);
			if (prop == NULL)
				continue;
#define DAX1_STR    "ORCL,sun4v-dax"
#define DAXPLUS_STR "ORCL,sun4v-dax-plus"
#define DAX2_STR    "ORCL,sun4v-dax2"
			if (strncmp(prop, DAX1_STR, strlen(DAX1_STR)))
				continue;
			printk(KERN_INFO "%s:   Found node 0x%Lx = %s\n", __func__, pn, prop);
			if (!strncmp(prop, DAXPLUS_STR, strlen(DAXPLUS_STR)))
				m7_bug = 0;
			else if (!strncmp(prop, DAX2_STR, strlen(DAX2_STR)))
				m7_bug = 0;
			else
				m7_bug = 1;
#if 0
			printk(KERN_INFO "%s:   MD indicates M7 chip %s bug.\n", __func__
			       m7_bug ? "with" : "without");
#endif
		}
	}

	if (sun4v_hvapi_register(HV_GRP_M7_DAX, 1, &stuff)) {
		printk(KERN_INFO "%s: hvapi_register failed\n", __func__);
		if ((force & 1) == 0)
			return -ENODEV;
	}
	else
		printk(KERN_INFO "%s: registered minor %ld\n", __func__, stuff);

	if ((ret = hv_get_hwqueue_size(&max_ccbs)) != 0) {
		printk(KERN_INFO "%s: get_hwqueue_size failed with status=%d and max_ccbs=%ld\n", __func__, ret, max_ccbs);
		if (force & 1)
			max_ccbs = 15;
		else
			return -ENODEV;
	}

	dax_hv_ccb_submit_maxlen = (uint32_t)(max_ccbs * sizeof (ccb_t));
	printk(KERN_INFO "%s: hwqueuesize = %ld, ccb_submit_maxlen = %d\n", __func__,
	       max_ccbs, dax_hv_ccb_submit_maxlen);
#ifndef U16_MAX
#define U16_MAX 65536
#endif
	if (max_ccbs == 0 || max_ccbs > U16_MAX) {
		printk(KERN_INFO " ...which is totally strange and unacceptable\n");
		if (force != 1)
			return -ENODEV;
	}

	if (dax_has_erratum_6008_numa()) {
		if (force == 2) {
			printk(KERN_INFO "%s: WARNING: pre-3.0 M7 CPU detected, but ignored due to FORCE option.\n", __func__);
			m7_bug = 0;
		} else {
			printk(KERN_INFO "%s: WARNING: pre-3.0 M7 CPU detected!\n", __func__);
			printk(KERN_INFO "%s: WARNING: Synthetic large page sizes rounded up to 64k or 4M.\n", __func__);
			m7_bug = 1;
		}
	} else {
		printk(KERN_INFO "%s: M7 3.0 CPU detected.\n", __func__);
		m7_bug = 0;
		if (force == 2) {
			printk(KERN_INFO "%s: but ignoring due to FORCE options.\n", __func__);
			printk(KERN_INFO "%s: WARNING: Synthetic large page sizes rounded up to 64k or 4M.\n", __func__);
			m7_bug = 1;
		}
	}

	if (alloc_chrdev_region(&first, 0, 1, "dax") < 0)
		return -ENXIO;

	cl = class_create(THIS_MODULE, "dax");
	if (cl == NULL) {
		printk(KERN_INFO "class_create failed\n");
		unregister_chrdev_region(first, 1);
		return -ENXIO;
	}

	if (device_create(cl, NULL, first, NULL, "dax") == NULL) {
		printk(KERN_INFO "device_create failed\n");
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -ENXIO;
	}

	cdev_init(&c_dev, &dax_fops);
	if (cdev_add(&c_dev, first, 1) == -1) {
		device_destroy(cl, first);
		printk(KERN_INFO "cdev_add failed\n");
		class_destroy(cl);
		unregister_chrdev_region(first, 1);
		return -ENXIO;
	}

	printk(KERN_INFO "Attached DAX module (built %s)\n", __TIMESTAMP__);

	return 0;
}

static void __exit
dax_detach(void)
{
	printk(KERN_INFO "Cleaning up DAX module.\n");
	if (!list_empty(&dax_mm_list)) {
		printk(KERN_INFO "%s: uh oh, there is stuff on the dax_mm_list\n", __func__);
	}
	printk(KERN_INFO "%s: dax_alloc_counter = %d\n", __func__, atomic_read(&dax_alloc_counter));
	printk(KERN_INFO "%s: dax_requested_mem = %dk\n", __func__, atomic_read(&dax_requested_mem));
	printk(KERN_INFO "%s: dax_actual_mem = %dk\n", __func__, atomic_read(&dax_actual_mem));
	printk(KERN_INFO "%s: dax_peak_waste = %dk\n", __func__, dax_peak_waste);
	cdev_del(&c_dev);
	device_destroy(cl, first);
	class_destroy(cl);
	unregister_chrdev_region(first, 1);
}
module_init(dax_attach);
module_exit(dax_detach);
MODULE_LICENSE("GPL");

/*
 * Logic of opens, closes, threads, contexts:
 *
 * open()/close()
 *
 * A thread may open the dax device as many times as it likes, but
 * each open must be bound to a separate thread before it can be used
 * to submit a transaction.
 *
 * The DAX_CCB_THR_INIT ioctl is called to create a context for the
 * calling thread and bind it to the file descriptor associated with
 * the ioctl. A thread must always use the fd to which it is bound.
 * A thread cannot bind to more than one fd, and one fd cannot be
 * bound to more than one thread.
 *
 * When a thread is finished, it should call the DAX_CCB_THR_FINI
 * ioctl to inform us that its context is no longer needed. This is
 * optional since close() will have the same effect for the context
 * associated with the fd being closed. However, if the thread dies
 * with its context still associated with the fd, then the fd cannot
 * ever be used again by another thread.
 *
 * The DAX_CA_DEQUEUE ioctl informs the driver that one or more
 * (contiguous) chunks of completion area buffers are no longer needed
 * and can be reused.
 *
 * The DAX_CCB_EXEC submits a coprocessor transaction using the
 * calling thread's context, which must match the context associated
 * with the associated fd.
 *
 */


/* ARGSUSED */
static int
dax_open(struct inode *i, struct file *f)
{
	DAX_DBG("%s();", __func__);
	return (0);
}

/* ARGSUSED */
static int
dax_close(struct inode *i, struct file *f)
{
	DAX_DBG("%s();", __func__);
	dax_state_destroy(f);
	return (0);
}

/* ARGSUSED */
static ssize_t
dax_read(struct file *f, char __user *data, size_t len, loff_t *off)
{
	char output[1024];
	static int size = 0;
	int requested = atomic_read(&dax_requested_mem);
	int granted   = atomic_read(&dax_actual_mem);

//	DAX_DBG("%s(len=0x%lx, off=%p, *off=0x%Lx);", __func__, len, off, *off);
	if (*off)
		return 0;
	size = sprintf(output, "memory use (Kb): requested %d, actual %d\n",
			requested, granted);
	if (copy_to_user(data, output, size))
		return -EFAULT;
	*off += size;
	return size;
}


/* ARGSUSED */
static long
dax_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
	DAX_DBG("%s: cmd=0x%x, f=%p, priv=%p", __func__, cmd, f, f->private_data);
	switch (cmd) {
	case DAXIOC_CCB_THR_INIT:
		return (dax_ioctl_ccb_thr_init( (void *)arg, f));
	case DAXIOC_CCB_THR_FINI:
		return (dax_ioctl_ccb_thr_fini(f));
	case DAXIOC_CA_DEQUEUE:
		return (dax_ioctl_ca_dequeue((void *)arg, f));
	case DAXIOC_CCB_EXEC:
		return (dax_ioctl_ccb_exec((void *)arg, f));
	case DAXIOC_VERSION:
		if (copy_to_user((void *)arg, &dax_version, sizeof(dax_version)))
			return -EFAULT;
		return 0;
	default:
		return perfcount_ioctl(f, cmd, arg);
	}
}

static void
dax_state_destroy(struct file *f)
{
	dax_ctx_t *ctx = (dax_ctx_t *) f->private_data;

	if (ctx != NULL) {
		dax_ctx_flush_decommit_ccbs(ctx);
		f->private_data = NULL;
		dax_ctx_fini(ctx);
	}
}

static int
dax_ioctl_ccb_thr_init(void *arg, struct file *f)
{
	dax_ccb_thr_init_arg_t usr_args;
	uint64_t compl_maplen;
	uint64_t compl_mapoff;
	dax_ctx_t *ctx;

	ctx = (dax_ctx_t *) f->private_data;

	/* Only one thread per open can create a context */
	if (ctx != NULL) {
		if (ctx->owner != current) {
			DAX_DBG("**** %s: This open already has an associated thread", __func__);
			return -EUSERS;
		} else {
			DAX_DBG("%s: duplicate CCB_THR_INIT ioctl", __func__);
			return -EINVAL;
		}
	}

	if (copy_from_user(&usr_args, (void *)arg, sizeof(usr_args)) != 0) {
		return -EFAULT;
	}

	DAX_DBG("%s: pid=%d, &ccb_maxlen = %p", __func__, current->pid, usr_args.dcti_ccb_buf_maxlen);

	compl_maplen = DAX_MMAP_SZ;
	compl_mapoff = DAX_MMAP_OFF;

	ASSERT(sizeof(dax_hv_ccb_submit_maxlen) == sizeof(*usr_args.dcti_ccb_buf_maxlen));
	if (copy_to_user(usr_args.dcti_ccb_buf_maxlen, &dax_hv_ccb_submit_maxlen, sizeof (dax_hv_ccb_submit_maxlen)) != 0) {
		DAX_DBG("%s: copyout dcti_ccb_buf_maxlen failed", __func__);
		return -EFAULT;
	}

	if (copy_to_user(usr_args.dcti_compl_maplen, &compl_maplen, sizeof (compl_maplen)) != 0) {
		DAX_DBG("%s: copyout dcti_compl_maplen failed", __func__);
		return -EFAULT;
	}

	if (copy_to_user(usr_args.dcti_compl_mapoff, &compl_mapoff, sizeof (compl_mapoff)) != 0) {
		DAX_DBG("%s: copyout dcti_compl_mapoff failed", __func__);
		return -EFAULT;
	}


	if ((ctx = dax_ctx_alloc()) == NULL) {
		DAX_DBG("%s: dax_ctx_alloc failed.", __func__);
		return -ENOMEM;
	}
	ctx->owner = current;
	f->private_data = ctx;
	return 0;
}

static int
dax_ioctl_ccb_thr_fini(struct file *f)
{
	dax_ctx_t *ctx = (dax_ctx_t *) f->private_data;

	if (ctx == NULL) {
		DAX_DBG("%s: CCB_THR_FINI ioctl called for open without previous CCB_THR_INIT ioctl", __func__);
		return -EINVAL;
	}

	if (ctx->owner != current) {
		DAX_DBG("%s: CCB_THR_FINI ioctl called from wrong thread ", __func__);
		return -EINVAL;
	}

#if 0
	/* wait for outstanding CCBs */
	dax_ctx_flush_decommit_ccbs(ctx);
	f->private_data = NULL;
	dax_ctx_fini(ctx);
#else
	dax_state_destroy(f);
#endif

	return (0);
}

static void
dax_overflow_check(dax_ctx_t *ctx, int idx)
{
	unsigned long output_size, input_size, virtp, page_size;
	ccb_hdr_t *hdr;
	ccb_t     *ccb;
	ccb_data_acc_ctl_t *access;
	struct vm_area_struct *vma;
	ccb_completion_area_t *ca = &ctx->ca_buf[idx];

	if (dax_debug == 0)
		return;

	if (ca->cmd_status != CCB_CMD_STAT_FAILED)
		return;

	if (ca->err_mask != CCB_CMD_ERR_POF)
		return;

	ccb = &ctx->ccb_buf[idx];
	hdr = CCB_HDR(ccb);

	access = (ccb_data_acc_ctl_t *) &ccb->dwords[QUERY_DWORD_DAC];
	output_size = access->output_buf_sz * 64 + 64;
	input_size  = access->input_cnt + 1;

	DAX_DBG("*************************");
	DAX_DBG("*DAX Page Overflow Report:");
	DAX_DBG("*  Output size requested = 0x%lx, output size produced = 0x%x",
		output_size, ca->output_sz);
	DAX_DBG("*  Input size requested = 0x%lx, input size processed = 0x%x",
		input_size, ca->n_processed);
	DAX_DBG("*  User virtual address analysis:");

	virtp = ccb->dwords[QUERY_DWORD_OUTPUT];

	if (hdr->at_dst == CCB_AT_RA) {
		DAX_DBG("*   Output address = 0x%lx physical, so no overflow possible", virtp);
	} else {
		/* output buffer was virtual, so page overflow is possible */
		page_size = 8*1024;
		if (hdr->at_dst == CCB_AT_VA_ALT) {
			vma = find_vma(current->mm, virtp);
			if (vma == NULL) {
				DAX_DBG("*   Output address = 0x%lx but is demapped, which precludes analysis", virtp);
			} else {
				page_size = vma_kernel_pagesize(vma);
			}
		}
		if (hdr->at_dst == CCB_AT_VA) {
			/* must be a synthetic page */
			if ((virtp & SVA_BASE_BITS) == SVA_BASE_64K)
				page_size = 64*1024;
			else if ((virtp & SVA_BASE_BITS) == SVA_BASE_4M)
				page_size = 4*1024*1024;
		}

		DAX_DBG("*   Output address = 0x%lx, page size = 0x%lx; page overflow %s",
			virtp, page_size,
			(virtp + ca->output_sz >= ((virtp + page_size) & ~(page_size - 1))) ? "LIKELY" : "UNLIKELY");
		DAX_DBG("*   Output size produced (0x%x) is %s the page bounds 0x%lx..0x%lx",
			ca->output_sz,
			(virtp + ca->output_sz >= ((virtp + page_size) & ~(page_size - 1))) ? "OUTSIDE" : "WITHIN",
			virtp, ((virtp + page_size) & ~(page_size - 1)));
	}

	virtp = ccb->dwords[QUERY_DWORD_INPUT];
	if (hdr->at_src0 == CCB_AT_VA_ALT) {
		vma = find_vma(current->mm, virtp);
		if (vma == NULL) {
			DAX_DBG("*   Input address = 0x%lx but is demapped, which precludes analysis", virtp);
		} else {
			page_size = vma_kernel_pagesize(vma);
			DAX_DBG("*   Input address = 0x%lx, page size = 0x%lx; page overflow %s",
				virtp, page_size,
				(virtp + input_size >= ((virtp + page_size) & ~(page_size - 1))) ? "LIKELY" : "UNLIKELY");
			DAX_DBG("*   Input size processed (0x%x) is %s the page bounds 0x%lx..0x%lx",
				ca->n_processed,
				(virtp + ca->n_processed >= ((virtp + page_size) & ~(page_size - 1))) ? "OUTSIDE" : "WITHIN",
				virtp, ((virtp + page_size) & ~(page_size - 1)));
		}
	} else
		DAX_DBG("*   Input address = 0x%lx physical, which can't happen", virtp);
	DAX_DBG("*************************");
}

static int
dax_ioctl_ca_dequeue(void *arg, struct file *f)
{
	dax_ctx_t *dax_ctx = (dax_ctx_t *) f->private_data;
	dax_ca_dequeue_arg_t usr_args;
	ccb_t *ccb_qhead;
	size_t ccb_contig_sz;
	uint32_t len_requested;
	uint32_t len_dequeued = 0;
	size_t decommit_len;
	size_t rv_len;
	int n_req_remain;
	int n_contig_avail;
	int n_contig_dq;
	int start_idx, end_idx;
	int i;
	int rv = 0;
	int n_ca_dequeue_contig_checks = 0;

	if (dax_ctx == NULL || dax_ctx->owner != current) {
		DAX_DBG("%s: CCB_INIT ioctl not previously called", __func__);
		return -EINVAL;
	}

	if (copy_from_user(&usr_args, (void *)arg, sizeof (usr_args)) != 0) {
		return -EFAULT;
	}

	DAX_DBG("%s: dcd_len_requested=%d", __func__, usr_args.dcd_len_requested);

	if (dax_validate_ca_dequeue_args(dax_ctx, &usr_args) != 0) {
		rv = -EINVAL;
		goto ca_dequeue_end;
	}

	/*
	 * The user length has been validated.
	 * If the kernel queue is empty, return EINVAL.
	 * Else, check that each CCB CA has completed in HW.
	 * If any CCB CA has not completed, return EBUSY.
	 */
	len_requested = usr_args.dcd_len_requested;
	n_req_remain = len_requested / sizeof (ccb_completion_area_t);
	len_dequeued = 0;

	/* can't we change to a do..while loop instead of using goto's? */
ca_dequeue_contig_block:
	ccb_qhead = dax_ccb_buffer_get_contig_block(dax_ctx, &ccb_contig_sz);

	BUG_ON(ccb_qhead == NULL);

	n_contig_avail = ccb_contig_sz / sizeof (ccb_t);
	start_idx = ccb_qhead - dax_ctx->ccb_buf;

	n_contig_dq = ((n_req_remain < n_contig_avail) ? n_req_remain :
	    n_contig_avail);

	end_idx = start_idx + n_contig_dq - 1;

	/* verify that all CAs have completed in HW */
	for (i = start_idx; i <= end_idx; i++) {
		uint8_t status;

		status = dax_ctx->ca_buf[i].cmd_status;

		if (status == CCB_CMD_STAT_NOT_COMPLETED) {
			DAX_DBG("%s: CCB completion area "
				"status == CCB_CMD_STAT_NOT_COMPLETED: "
				"fail request to free completion index=%d",
				__func__, i);
			rv = -EBUSY;
			goto ca_dequeue_end;
		}

		dax_overflow_check(dax_ctx, i);

		/* free any locked page associated with this ccb */
		if (dax_ctx->pages[i]) {
			set_page_dirty(dax_ctx->pages[i]);
			put_page(dax_ctx->pages[i]);
			DAX_DBG("%s: freeing page %p", __func__, dax_ctx->pages[i]);
			dax_ctx->pages[i] = NULL;
		}
		else {
			ccb_t *ccb = &dax_ctx->ccb_buf[i];
			ccb_hdr_t *hdr = CCB_HDR(ccb);
			if (hdr->at_dst == CCB_AT_VA_ALT) {
				DAX_DBG("%s: why isn't page for 0x%Lx locked?", __func__, ccb->dwords[QUERY_DWORD_OUTPUT]);
			}
		}

		if (IS_LONG_CCB(&dax_ctx->ccb_buf[i])) {
			/*
			 * Validate that the user must dequeue 2 CAs for a long
			 * CCB.  In other words, the last entry in a contig
			 * block cannot be a long CCB.
			 */
			if (i == end_idx) {
				rv = -EINVAL;
				DAX_DBG("%s: invalid attempt to dequeue single CA for long CCB, index=%d",
					__func__, i);
				goto ca_dequeue_end;
			} else {
				/* skip over 64B data of long CCB */
				i++;
			}
		}
	}

	/* Free buffer. Update accounting. */
	decommit_len = n_contig_dq * sizeof (ccb_t);
	rv_len = dax_ccb_buffer_decommit(dax_ctx, decommit_len);

	ASSERT(rv_len == decommit_len);

	len_dequeued += (n_contig_dq * sizeof (ccb_completion_area_t));
	n_req_remain -= n_contig_dq;

	if (n_req_remain > 0) {
		DAX_DBG("%s: checking additional ccb_buffer "
			"contig block, n_req_remain=%d", __func__, n_req_remain);
		n_ca_dequeue_contig_checks++;
		ASSERT(n_ca_dequeue_contig_checks == 1);
		goto ca_dequeue_contig_block;
	}

ca_dequeue_end:
	DAX_DBG("%s: copyout len_dequeued=%d", __func__, len_dequeued);

	if (copy_to_user(usr_args.dcd_len_dequeued, &len_dequeued, sizeof (len_dequeued)) != 0) {
		DAX_DBG("%s: copyout dcd_len_dequeued failed", __func__);
		return -(EFAULT);
	}

	return (rv);
}

static int
dax_validate_ca_dequeue_args(dax_ctx_t *dax_ctx, dax_ca_dequeue_arg_t *usr_args)
{
	/* requested len must be multiple of completion area size */
	if ((usr_args->dcd_len_requested % sizeof (ccb_completion_area_t)) != 0) {
		DAX_DBG("%s: dequeue len (%d) not 128B aligned",
			__func__, usr_args->dcd_len_requested);
		return -1;
	}

	/* and not more than current buffer entry count */
	if ((usr_args->dcd_len_requested / sizeof (ccb_completion_area_t)) >
	    (dax_ctx->bufcnt / sizeof (ccb_t))) {
		DAX_DBG("%s: dequeue len (%d bytes, %ld CAs) "
			"more than current CA buffer count (%ld CAs)", __func__,
			usr_args->dcd_len_requested,
			(usr_args->dcd_len_requested /
			 sizeof (ccb_completion_area_t)),
			dax_ctx->bufcnt / sizeof (ccb_t));
		return -1;
	}

	/* reject zero length */
	if (usr_args->dcd_len_requested == 0)
		return -1;

	return 0;
}

static char *dax_vm_print(char *buf, dax_vma_t *dv)
{
	int len;

	len = sprintf(buf, "vma %p, kva=%p, uva=0x%lx, pa=0x%lx",
		dv->vma, dv->kva, dv->vma ? dv->vma->vm_start : 0, dv->pa);
	len += sprintf(buf+len, "\n\trequested_length=0x%lx, actual_length=0x%lx",
		       dv->requested_length, dv->actual_length);
	if (dv->sva)
		sprintf(buf+len, "\n\tsva=0x%lx, tte=0x%lx", dv->sva, dv->tte);
	return buf;
}

static void dax_vm_close(struct vm_area_struct *vma)
{
	dax_vma_t *dv;
	dax_mm_t  *dax_mm;
	char msg[200];

	dv = vma->vm_private_data;
	DAX_DBG_MAP("%s: vma=%p, dv=%p", __func__, vma, dv);
	if (dv == NULL) {
		printk(KERN_ALERT "%s: dv NULL in dax_vm_close!! Impossible!!\n", __func__);
		return;
	}
	if (dv->vma != vma) {
		DAX_DBG_MAP("%s: munmap(0x%lx, 0x%lx) differs from mmap length 0x%lx",
			    __func__, vma->vm_start, vma->vm_end - vma->vm_start, dv->requested_length);
		return;
	}

	dax_mm = dv->dax_mm;
	if (dax_mm == NULL) {
		printk(KERN_ALERT "%s: dv->dax_mm NULL in dax_vm_close!! Impossible!!\n", __func__);
		return;
	}

	DAX_DBG_MAP("%s: freeing %s", __func__, dax_vm_print(msg, dv));
	spin_lock(&dax_mm->lock);
	vma->vm_private_data = NULL;

	dv->vma = NULL;   /* signifies no mapping exists and prevents new transactions */
	dax_vma_drain(dv);

	kfree(dv->kva);
	list_del_init(&dv->vma_list);
	atomic_add(-dv->requested_length / 1024, &dax_requested_mem);
	atomic_add(-dv->actual_length / 1024, &dax_actual_mem);
	kfree(dv);
	dax_mm->vma_count--;
	atomic_dec(&dax_alloc_counter);

	/* if vma_list is empty and ctx list is empty, clean up this dax_mm_t */
	spin_lock(&dax_mm_lock);
	if (list_empty(&dax_mm->ctx_list) && list_empty(&dax_mm->vma_list)) {
		list_del(&dax_mm->mm_list);
		DAX_DBG_LIST("%s: freeing dax_mm with vma_count=%d, ctx_count=%d",
			     __func__, dax_mm->vma_count, dax_mm->ctx_count);
		kfree(dax_mm);
	}
	else
		spin_unlock(&dax_mm->lock);
	spin_unlock(&dax_mm_lock);
}


void data_snoop(dax_vma_t *dv)
{
	unsigned long data, *datap;

	datap = dv->kva;
	data = *datap;
	if (dv->vma == NULL) {
		DAX_DBG_MAP("%s: vma unmapped *kva=0x%lx", __func__, data);
		return;
	}

	DAX_DBG_MAP("%s: *kva=0x%lx", __func__, data);

	datap = (unsigned long *) dv->vma->vm_start;
	if (copy_from_user(&data, datap, sizeof(unsigned long))) {
		DAX_DBG_MAP("%s: efault in copy_from_user(%p)", __func__, datap);
	} else {
		DAX_DBG_MAP("%s: *uva=0x%lx", __func__, data);
	}

	data = 0;
	datap = (unsigned long *) dv->pa;
	__asm__ __volatile__("ldxa [%1] #ASI_PHYS_USE_EC, %0\n\t" : "=r" (data) : "r" (datap) :);
	DAX_DBG_MAP("%s: *pa=0x%lx", __func__, data);
	DAX_DBG_MAP("%s: vma->vm_mm=%p, current->mm=%p",
		    __func__, dv->vma->vm_mm, current->mm);
}

/*
 * open function called if the vma is split;
 * usually happens in response to a partial munmap()
 */
void dax_vm_open(struct vm_area_struct *vma)
{
	DAX_DBG_MAP("%s: call with va=0x%lx, len=0x%lx",
	       __func__, vma->vm_start, vma->vm_end - vma->vm_start);
	DAX_DBG_MAP("%s: prot=0x%lx, flags=0x%lx",
	       __func__, pgprot_val(vma->vm_page_prot), vma->vm_flags);
//	dump_stack();
}

/*
 * fault, mkwrite, access, name: should never be called
 */
int dax_vm_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	DAX_DBG_MAP("%s: call with va=0x%lx, len=0x%lx",
	       __func__, vma->vm_start, vma->vm_end - vma->vm_start);
	if (dax_debug & DAX_DBG_FLG_MAP)
		dump_stack();
	return 0;
}

int dax_vm_page_mkwrite(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	DAX_DBG_MAP("%s: call with va=0x%lx, len=0x%lx",
	       __func__, vma->vm_start, vma->vm_end - vma->vm_start);
	if (dax_debug & DAX_DBG_FLG_MAP)
		dump_stack();
	return 0;
}

int dax_vm_access(struct vm_area_struct *vma, unsigned long addr,
		  void *buf, int len, int write)
{
	DAX_DBG_MAP("%s: call with va=0x%lx, len=0x%lx",
	       __func__, vma->vm_start, vma->vm_end - vma->vm_start);
	if (dax_debug & DAX_DBG_FLG_MAP)
		dump_stack();
	return 0;
}

const char *dax_vm_name(struct vm_area_struct *vma)
{
	DAX_DBG_MAP("%s: call with va=0x%lx, len=0x%lxn",
	       __func__, vma->vm_start, vma->vm_end - vma->vm_start);
	return "DAX";
}

static const struct vm_operations_struct dax_vm_ops = {
	.open  = dax_vm_open,
	.close = dax_vm_close,
	.fault = dax_vm_fault,
	.page_mkwrite = dax_vm_page_mkwrite,
	.access = dax_vm_access,
	.name  = dax_vm_name
};

void fizzz(dax_vma_t *dv)
{
	int len;
	unsigned long addr, val = 0xbbbbbbbb77777777;

	for (addr = dv->pa, len = 0; len < dv->actual_length; len = len + 8) {
		asm volatile("stxa  %0, [%1+%2] 0x14\n\t" : : "r" (val), "r" (addr), "r" (len));
	}
}

/* mmu_map_addr()
 * TRAP:	HV_MMU_MAP_ADDR_TRAP
 * ARG0:	virtual address
 * ARG1:	mmu context
 * ARG2:	TTE
 * ARG3:	flags (HV_MMU_{IMMU,DMMU})
 * ERRORS:	EINVAL		Invalid virtual address, mmu context, or flags
 *		EBADPGSZ	Invalid page size value
 *		ENORADDR	Invalid real address in TTE
 *
 * Create a non-permanent mapping using the given TTE, virtual
 * address, and mmu context.  The flags argument determines which
 * (data, or instruction, or both) TLB the mapping gets loaded into.
 *
 * The behavior is undefined if the valid bit is clear in the TTE.
 *
 * Note: This API call is for privileged code to specify temporary translation
 *       mappings without the need to create and manage a TSB.
 */
long mmu_map_addr(long va, long ctx, long tte, long flags)
{
	long status;
	__asm__ __volatile__(
		"mov %[va], %%o0\n\t"
		"clr %%o1\n\t"
		"mov %[tte], %%o2\n\t"
		"mov %[flags], %%o3\n\t"
		"ta %[trap]\n\t"
		"mov %%o0, %[status]\n"
		: [status] "=r" (status)
		: [va]     "r" (va),
		  [tte]    "r" (tte),
		  [flags]  "r" (flags),
		  [trap]   "i" (HV_MMU_MAP_ADDR_TRAP)
		: "o0", "o1", "o2", "o3"
		);
	return status;
}

static int
dax_alloc_ram(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long pa, page;
	char *kva;
	dax_vma_t *dv;
	size_t len;
	char msg[200];
	int ret, offset, waste;
	dax_ctx_t *dax_ctx = (dax_ctx_t *) filp->private_data;

	len = vma->vm_end - vma->vm_start;
	if (len & (PAGE_SIZE - 1)) {
		DAX_DBG_MAP("%s: request (0x%lx) not a multiple of page size", __func__, len);
		return -ENOMEM;
	}
	if (m7_bug) {
		if (len <= 64*1024)
			len = 64*1024;
		else if (len > 4*1024*1024) {
			DAX_DBG_MAP("%s: unsupported length 0x%lx exceeds 4M virtual page size",
				    __func__, len);
			return -ENOMEM;
		}
		else
			len = 4*1024*1024;
	}

	DAX_DBG_MAP("%s: requested length=0x%lx, actual length=0x%lx", __func__, vma->vm_end - vma->vm_start, len);

	if (dax_ctx->dax_mm == NULL) {
		DAX_DBG_MAP("%s: no dax_mm for ctx %p!", __func__, dax_ctx);
		return -ENOMEM;
	}

	kva = kmalloc(len, GFP_KERNEL);
	if (kva == NULL)
	{
		DAX_DBG_MAP("%s: kmalloc failed", __func__);
		return -ENOMEM;
	}

	if ((unsigned long)kva & (PAGE_SIZE - 1)) {
		DAX_DBG_MAP("%s: kmalloc returned unaligned (8k) addr %p", __func__, kva);
		kfree(kva);
		return -ENOMEM;
	}
	if (m7_bug)
		if ((unsigned long)kva & (len - 1)) {
			DAX_DBG_MAP("%s: kmalloc returned unaligned (%ldk) addr %p", __func__, len/1024, kva);
			kfree(kva);
			return -ENOMEM;

		}

	memset(kva, 0, len);	/* prolly no need for this */

	pa = virt_to_phys((void *)kva);
	for (offset = 0; offset < (vma->vm_end - vma->vm_start); offset += PAGE_SIZE) {

		/* while we're here, verify physical contiguity */
		if (virt_to_phys(kva + offset) != pa + offset) {
			DAX_DBG_MAP("%s: discontinuity detected in supposedly physical pages:\n"
				    "  virt_to_phys(%p+0x%x) 0x%lx != virt_to_phys(%p) + 0x%x, instead it's 0x%lx",
				    __func__, kva, offset, virt_to_phys(kva + offset), kva, offset, pa + offset);
			return -EIO;
		}

		page = virt_to_phys(kva + offset) >> PAGE_SHIFT;
		if ( (ret = remap_pfn_range(vma, vma->vm_start + offset, page, PAGE_SIZE, vma->vm_page_prot)) != 0 ) {
			DAX_DBG_MAP("%s: remap failed with error %d for uva 0x%lx, page 0x%lx", __func__, ret, vma->vm_start + offset, page);
			return ret;
		}
	}
	DAX_DBG_MAP("%s: mapped kva 0x%lx = uva 0x%lx to pa 0x%lx", __func__,
		    (unsigned long) kva, vma->vm_start, pa);

	dv = kzalloc(sizeof(*dv), GFP_KERNEL);
	dv->vma = vma;
	dv->kva = kva;
	dv->pa = pa;
	dv->requested_length = vma->vm_end - vma->vm_start;
	dv->actual_length = len;
	dv->dax_mm = dax_ctx->dax_mm;
	INIT_LIST_HEAD(&dv->vma_list);

	spin_lock(&dax_ctx->dax_mm->lock);
	list_add(&dv->vma_list, &dax_ctx->dax_mm->vma_list);
	dax_ctx->dax_mm->vma_count++;
	spin_unlock(&dax_ctx->dax_mm->lock);
	atomic_inc(&dax_alloc_counter);
	atomic_add(dv->requested_length / 1024, &dax_requested_mem);
	atomic_add(dv->actual_length / 1024, &dax_actual_mem);
	waste = atomic_read(&dax_actual_mem) - atomic_read(&dax_requested_mem);
	if (waste > dax_peak_waste)
		dax_peak_waste = waste;

	vma->vm_ops = &dax_vm_ops;
	vma->vm_private_data = dv;

	DAX_DBG_MAP("%s: mapped %s", __func__, dax_vm_print(msg, dv));
#if 0
	DAX_DBG_MAP("%s: prot=0x%lx, flags=0x%lx",
		    __func__, pgprot_val(vma->vm_page_prot), vma->vm_flags);
	*(unsigned long *)va = 0x1234567890abcdef;
	data_snoop(dv);
#endif
	if (dax_debug & DAX_DBG_FLG_MAP)
		fizzz(dv);

	if (m7_bug) {
		/*
		 * Virtual address selection
		 *
		 * We need a new, temporary virtual address to use for
		 * our synthetic page.  Properly, we ought to call
		 * __get_vm_area() to allocate a VA range to use.  I
		 * chose not to do this, and considered many factors
		 * in this decision:
		 *
		 *  1) Reserving a VA range with __get_vm_area() means
		 *     that at some later time, we need to release it
		 *     with a call to free_vmap_area(), which is not
		 *     exported. So it is not clear how to properly
		 *     release the area.  free_vm_area() may work, but
		 *     it assumes that the area is mapped, and tries
		 *     to unmap it. Also, when to do this release?
		 *     Maybe when the user segment is unmapped?
		 *
		 *  2) We need to be able to tell the page size by
		 *     looking at just the VA.  This is for page
		 *     overflow detection. So encoding the page size
		 *     right in the VA is an easy way to accomplish
		 *     this.
		 *
		 *  3) Since an RA is 50 bits, and VA is 54 bits on
		 *     M7, we've got 4 bits to play with, and we can
		 *     keep the entire RA in the low order 50 bits of
		 *     the VA, which makes for a trivial way to ensure
		 *     a unique VA for any RA without any fancy data
		 *     structure and algorithms to keep track of VA
		 *     allocations.
		 *
		 *  4) This code is a temporary workaround for a chip
		 *     bug, and so in the near future, it should not
		 *     be activated, so there is no point in
		 *     engineering a perfect long term solution for
		 *     this case.
		 *
		 *  5) The only disadvantage is that we are "claiming"
		 *     a range of virtual addresses, and some future
		 *     use of this range could conflict.
		 *
		 *  6) An upcoming change to the DAX hypervisor API
		 *     will allow us to specify the page size along
		 *     with an RA. This enhancement will greatly
		 *     simplify the code, and we will no longer need
		 *     this synthetic virtual address range at all.
		 *
		 * 
		 */
		if (len == 64*1024) {
			dv->tte = _PAGE_SZ64K_4V | _PAGE_VALID | _PAGE_CP_4V | _PAGE_P_4V | _PAGE_W_4V;
			dv->sva = (pa & ~SVA_BASE_BITS) | SVA_BASE_64K;
		} else {
			dv->tte = _PAGE_SZ4MB_4V | _PAGE_VALID | _PAGE_CP_4V | _PAGE_P_4V | _PAGE_W_4V;
			dv->sva = (pa & ~SVA_BASE_BITS) | SVA_BASE_4M;
		}

		dv->tte |= dv->pa & _PAGE_PADDR_4V;
#if 0
		if (dax_debug & DAX_DBG_FLG_MAP) {
			long status, junk;
			int i;

			status = mmu_map_addr(dv->sva, 0, dv->tte, HV_MMU_DMMU);
			DAX_DBG_MAP("%s: mmu_map_addr(0x%lx, 0, 0x%lx, 0x%x) returned 0x%lx",
				    __func__, dv->sva, dv->tte, HV_MMU_DMMU, status);
			for (i=0; i<len; i=i+1024) {
				*(unsigned long *)(dv->kva + i) = 0xdeadbeef;
				status = probe_kernel_read(&junk, (void *)dv->sva + i, sizeof(long));
				if (junk != 0xdeadbeef)
					DAX_DBG_MAP("%s: data mismatch for 0x%lx + 0x%x", __func__, dv->sva, i);
				if (status != 0)
					DAX_DBG_MAP("%s: probe failed for 0x%lx + 0x%x", __func__, dv->sva, i);
			}
			status = probe_kernel_read(&junk, (void *)dv->sva + len, sizeof(long));
			if (status != 0) {
				DAX_DBG_MAP("%s: probe failed for 0x%lx + 0x%lx (expected)", __func__, dv->sva, len);
			} else {
				DAX_DBG_MAP("%s: probe didn't fail for 0x%lx + 0x%lx (unexpected)", __func__, dv->sva, len);
			}
		}
#endif
	}

	return 0;
}


/*
 * The DAX completion area array contains status information about
 * each transaction. It is written directly by the coprocessor
 * and read directly by the user via an mmap region.
 */
static int
dax_devmap(struct file *f, struct vm_area_struct *vma)
{
	unsigned long page, pos;
	unsigned long start;
	dax_ctx_t *dax_ctx = (dax_ctx_t *) f->private_data;
        size_t len = vma->vm_end - vma->vm_start;

	DAX_DBG("%s: len=0x%lx, flags=0x%lx", __func__, len, vma->vm_flags);

	if (dax_ctx == NULL) {
		DAX_DBG("%s: CCB_INIT ioctl not previously called", __func__);
		return -EINVAL;
	}
	if (dax_ctx->owner != current) {
		DAX_DBG("%s: devmap called from wrong thread", __func__);
		return -EINVAL;
	}

	if (vma->vm_flags & VM_WRITE)
		return dax_alloc_ram(f, vma);

	if (len != dax_ctx->ca_buflen) {
		DAX_DBG_MAP("%s: len(%lu) != dax_ctx->ca_buflen(%u)", __func__, len, dax_ctx->ca_buflen);
		return -EINVAL;
	}

	pos = (unsigned long) dax_ctx->ca_buf;
	for (start = vma->vm_start;
	     len > 0;
	     start += PAGE_SIZE, pos += PAGE_SIZE, len -= PAGE_SIZE) {
                page = virt_to_phys((void *)pos) >> PAGE_SHIFT;
                if (remap_pfn_range(vma, start, page, PAGE_SIZE, vma->vm_page_prot))
                        return -EAGAIN;
        }
	DAX_DBG_MAP("%s: mmapped completion area at uva 0x%lx", __func__, vma->vm_start);
	return (0);
}

static dax_ctx_t *
dax_ctx_alloc(void)
{
	dax_ctx_t *dax_ctx;
	dax_mm_t *dm;
	struct list_head *p;

	dax_ctx = kzalloc(sizeof(dax_ctx_t), GFP_KERNEL);

	/* allocate CCB buffer */
	ASSERT((DAX_CCB_BUF_SZ & (DAX_CCB_BUF_SZ - 1)) == 0);

	dax_ctx->ccb_buf = kmalloc(DAX_CCB_BUF_SZ, GFP_KERNEL);
	ASSERT(dax_ctx->ccb_buf != NULL);
	dax_ctx->ccb_buf_ra = virt_to_phys(dax_ctx->ccb_buf);
	dax_ctx->ccb_buflen = DAX_CCB_BUF_SZ;
	dax_ctx->ccb_submit_maxlen = dax_hv_ccb_submit_maxlen;

	DAX_DBG("%s: state->dax_ctx->ccb_buf=0x%p, "
		"ccb_buf_ra=0x%Lx, ccb_buflen=%d", __func__, (void *)dax_ctx->ccb_buf,
		dax_ctx->ccb_buf_ra, dax_ctx->ccb_buflen);

	/* allocate CCB completion area buffer */
	ASSERT((DAX_CA_BUF_SZ & (DAX_CA_BUF_SZ - 1)) == 0);
	dax_ctx->ca_buf = kmalloc(DAX_CA_BUF_SZ, GFP_KERNEL);
	ASSERT(dax_ctx->ca_buf != NULL);
	dax_ctx->ca_buflen = DAX_CA_BUF_SZ;
	dax_ctx->ca_buf_ra = virt_to_phys(dax_ctx->ca_buf);
	DAX_DBG("%s: allocated 0x%x bytes for ca_buf", __func__, dax_ctx->ca_buflen);
	memset(dax_ctx->ca_buf, 0, DAX_CA_BUF_SZ);

	/* allocate page array */
	dax_ctx->pages = kzalloc(DAX_CCB_BUF_NELEMS * sizeof(struct page *), GFP_KERNEL);
	ASSERT(dax_ctx->pages != NULL);

	/* initialize buffer accounting */
	dax_ctx->a_start = 0;
	dax_ctx->a_end = 0;
	dax_ctx->b_end = 0;
	dax_ctx->resv_start = 0;
	dax_ctx->resv_end = 0;
	dax_ctx->bufcnt = 0;
	dax_ctx->ccb_count = 0;
	dax_ctx->fail_count = 0;

	DAX_DBG("%s: dax_ctx=0x%p, dax_ctx->ca_buf=0x%p, "
		"ca_buf_ra=0x%Lx, ca_buflen=%d, ",
		__func__, (void *)dax_ctx, (void *)dax_ctx->ca_buf,
		dax_ctx->ca_buf_ra, dax_ctx->ca_buflen);

	/* look for existing mm context */
	spin_lock(&dax_mm_lock);
	list_for_each(p, &dax_mm_list) {
		dm = list_entry(p, dax_mm_t, mm_list);
		if (dm->this_mm == current->mm) {
			dax_ctx->dax_mm = dm;
			DAX_DBG_MAP("%s: existing dax_mm found: %p", __func__, dm);
			break;
		}
	}

	/* did not find an existing one, must create it */
	if (dax_ctx->dax_mm == NULL) {
		dm = kmalloc(sizeof(*dm), GFP_KERNEL);
		INIT_LIST_HEAD(&dm->mm_list);
		INIT_LIST_HEAD(&dm->vma_list);
		INIT_LIST_HEAD(&dm->ctx_list);
		spin_lock_init(&dm->lock);
		dm->this_mm = current->mm;
		dm->vma_count = 0;
		dm->ctx_count = 0;
		list_add(&dm->mm_list, &dax_mm_list);
		dax_ctx->dax_mm = dm;
		DAX_DBG_MAP("%s: no dax_mm found, creating and adding to dax_mm_list: %p", __func__, dm);
	}
	spin_unlock(&dax_mm_lock);
	/* now add this ctx to the list of thread for this mm context */
	INIT_LIST_HEAD(&dax_ctx->ctx_list);
	spin_lock(&dm->lock);
	list_add(&dax_ctx->ctx_list, &dax_ctx->dax_mm->ctx_list);
	dax_ctx->dax_mm->ctx_count++;
	spin_unlock(&dm->lock);

	DAX_DBG("%s: allocated ctx %p", __func__, dax_ctx);
	return (dax_ctx);
}

static void
dax_ctx_fini(dax_ctx_t *ctx)
{
	int i;
	dax_mm_t *dax_mm;

	ASSERT(ctx != NULL);

	kfree(ctx->ccb_buf);
	ctx->ccb_buf = NULL;

	kfree(ctx->ca_buf);
	ctx->ca_buf = NULL;

	for (i=0; i<DAX_CCB_BUF_NELEMS; i++)
//		ASSERT(ctx->pages[i] == NULL);
		if (ctx->pages[i] != NULL)
			DAX_DBG_DRV("%s: still not freed pages[%d] = %p", __func__, i, ctx->pages[i]);


	kfree(ctx->pages);
	ctx->pages = NULL;

	dax_mm = ctx->dax_mm;
	if (dax_mm == NULL) {
		DAX_DBG_MAP("%s: No dax_mm!!", __func__);
	} else {
		spin_lock(&dax_mm->lock);
		list_del(&ctx->ctx_list);
		/* if vma_list is empty and ctx list is empty, clean up this dax_mm_t */
		spin_lock(&dax_mm_lock);
		if (list_empty(&dax_mm->ctx_list) && list_empty(&dax_mm->vma_list)) {
			list_del(&dax_mm->mm_list);
			DAX_DBG_LIST("%s: freeing dax_mm with vma_count=%d, ctx_count=%d",
				     __func__, dax_mm->vma_count, dax_mm->ctx_count);
			kfree(dax_mm);
		}
		else
			spin_unlock(&dax_mm->lock);
		spin_unlock(&dax_mm_lock);
	}

	DAX_DBG_DRV("%s: CCB count: %d good, %d failed", __func__, ctx->ccb_count, ctx->fail_count);
	kfree(ctx);
}

static int
dax_validate_ccb_exec_args(dax_ccb_exec_arg_t *usr_args)
{
	switch (usr_args->dce_hwq_type) {
	case DAX_CCB_HWQ_TYPE_QUERY:
	case DAX_CCB_HWQ_TYPE_MSG:
		break;
	default:
		DAX_DBG("%s: invalid ccb_hwq_type(%d)", __func__,
			usr_args->dce_hwq_type);
		return -ENOSYS;
	}

	if (usr_args->dce_ccb_buf_len > dax_hv_ccb_submit_maxlen ||
	    (usr_args->dce_ccb_buf_len % sizeof (ccb_t)) != 0 ||
	    usr_args->dce_ccb_buf_len == 0) {
		DAX_DBG("%s: invalid usr_args->dce_ccb_len(%d)",
			__func__, usr_args->dce_ccb_buf_len);
		return -ERANGE;
	}

	DAX_DBG("%s: usr_args->dce_hwq_type=%d, "
		"usr_args->dce_ccb_buf_len=%d", __func__, usr_args->dce_hwq_type,
		usr_args->dce_ccb_buf_len);

	return 0;
}

static int
dax_validate_ccb(ccb_t *ccb)
{
	ccb_hdr_t *hdr = CCB_HDR(ccb);

	/*
	 * The user is not allowed to specify real address types
	 * in the CCB header.  This must be enforced by the kernel
	 * before submitting the CCBs to HV.
	 *
	 * The allowed values are:
	 *	hdr->at_dst	VA/RKEY/IMM only
	 *	hdr->at_src0	VA/RKEY/IMM only
	 *	hdr->at_src1	VA/RKEY/IMM only
	 *	hdr->at_tbl	VA/IMM only
	 *
	 * Note: RKEY is not valid for query CCB and IMM is only
	 * valid for certain opcodes, but the kernel is not validating
	 * at this level of granularity.  The HW will flag invalid
	 * address types.  The required check is that the user must
	 * not be allowed to specify real address types.
	 */

	if (!(hdr->at_dst == CCB_AT_VA || hdr->at_dst == CCB_AT_RKEY ||
	    hdr->at_dst == CCB_AT_IMM)) {
		DAX_DBG("%s: invalid at_dst address type (%d) "
			"in user CCB", __func__, hdr->at_dst);
		return -EINVAL;
	}

	if (!(hdr->at_src0 == CCB_AT_VA || hdr->at_src0 == CCB_AT_RKEY ||
	    hdr->at_src0 == CCB_AT_IMM)) {
		DAX_DBG("%s: invalid at_src0 address type (%d) "
			"in user CCB", __func__, hdr->at_src0);
		return -EINVAL;
	}

	if (!(hdr->at_src1 == CCB_AT_VA || hdr->at_src1 == CCB_AT_RKEY ||
	    hdr->at_src1 == CCB_AT_IMM)) {
		DAX_DBG("%s: invalid at_src1 address type (%d) "
			"in user CCB", __func__, hdr->at_src1);
		return -EINVAL;
	}

	if (!(hdr->at_tbl == CCB_AT_VA || hdr->at_tbl == CCB_AT_IMM)) {
		DAX_DBG("%s: invalid at_tbl address type (%d) "
			"in user CCB", __func__, hdr->at_tbl);
		return -EINVAL;
	}

	return (0);
}

static void
dax_prt_ccbs(ccb_t *ccb, uint64_t len)
{
	int nelem = len / sizeof (ccb_t);
	int i, j;

	ASSERT(len % sizeof (ccb_t) == 0);

	DAX_DBG("%s: ccb buffer (processed):", __func__);
	for (i = 0; i < nelem; i++) {
		DAX_DBG("%s: ccb[%d]", __func__, i);
		for (j = 0; j < DWORDS_PER_CCB; j++) {
			DAX_DBG("%s:\tccb[%d].dwords[%d]= "
				"0x%Lx", __func__, i, j, ccb[i].dwords[j]);
		}
	}
}

static int
dax_ioctl_ccb_exec(void *arg, struct file *f)
{
	dax_ccb_exec_arg_t usr_args;
	ccb_exec_ctl_t exec_ctl = {0};
	dax_ctx_t *dax_ctx = (dax_ctx_t *) f->private_data;
	ccb_t *ccb_buf;
	size_t nreserved;
	uint64_t ca_region_off;
	int rv, hv_rv;
	uint64_t current_time;

	asm volatile("rd  %%tick, %0\nmov %0, %0\n" : "=r" (current_time));
	if (dax_ctx == NULL) {
		DAX_DBG("%s: CCB_INIT ioctl not previously called", __func__);
		return -ENOENT;
	}
	if (dax_ctx->owner != current) {
		DAX_DBG("%s: wrong thread", __func__);
		return -EUSERS;
	}

	if (copy_from_user(&usr_args, (void *)arg, sizeof (usr_args)) != 0) {
		DAX_DBG("%s: copyin of user args failed", __func__);
		return -EFAULT;
	}

#if 0
	printk(KERN_INFO "%s: time measurement #1 = %Ld cycles\n", __func__,
	       current_time - usr_args.dce_timer);
#endif

	DAX_DBG("%s args: hwq_type = %d, ccb_buf_len=%d, buf_addr=%p, "
		"&submitted_buf_len=%p, "
		"&ca_region_off=%p, "
		"&ccb_status=%p",
		__func__,
		usr_args.dce_hwq_type, usr_args.dce_ccb_buf_len, usr_args.dce_ccb_buf_addr,
		usr_args.dce_submitted_ccb_buf_len, usr_args.dce_ca_region_off,
		usr_args.dce_ccb_status);

	if ( (rv = dax_validate_ccb_exec_args(&usr_args)) != 0)
		return rv;

	/*
	 * Could get rid of this from the API and use the opcode of the first
	 * CCB to determine the target HW queue.  A mix of queue types will
	 * either fail in HV or HW (should be graceful).
	 */
	exec_ctl.hwq_type = usr_args.dce_hwq_type;

	/* Check for available buffer space. */
	ccb_buf = dax_ccb_buffer_reserve(dax_ctx, usr_args.dce_ccb_buf_len, &nreserved);
//	printk(KERN_INFO "reserved address %p for ccb_buf\n", ccb_buf);

	/*
	 * We don't attempt a partial submission since that would require extra
	 * logic to not split a long CCB at the end.  This would be an
	 * enhancement.
	 */
	if (ccb_buf == NULL || nreserved != usr_args.dce_ccb_buf_len) {
		DAX_DBG("%s: insufficient kernel CCB resources: "
			"user needs to free completion area space and retry",
			__func__);
		return -ENOBUFS;
	}

	/*
	 * Copy user CCBs.  Here we copy the entire user buffer and later
	 * validate the contents by running the buffer.
	 */
	if (copy_from_user(ccb_buf, (void *)usr_args.dce_ccb_buf_addr, usr_args.dce_ccb_buf_len) != 0) {
		DAX_DBG("%s: copyin of user CCB buffer failed", __func__);
		return -EFAULT;
	}

	rv = dax_preprocess_usr_ccbs(dax_ctx, ccb_buf, usr_args.dce_ccb_buf_len);

	if (rv != 0)
		return rv;

	dax_map_segment(dax_ctx, ccb_buf, usr_args.dce_ccb_buf_len);

	rv = dax_lock_pages(dax_ctx, ccb_buf, usr_args.dce_ccb_buf_len);
	if (rv != 0)
		return rv;

	/* hypercall */
	hv_rv = dax_ccb_hv_submit(dax_ctx, ccb_buf, usr_args.dce_ccb_buf_len, &exec_ctl);

	/* Update based on actual number of submitted CCBs. */
	if (hv_rv == 0) {
		dax_ccb_buffer_commit(dax_ctx, exec_ctl.submitted_ccb_buf_len);
		dax_ctx->ccb_count++;
	} else {
		dax_ctx->fail_count++;
		DAX_DBG("%s: submit failed, status=%d, nomap=0x%Lx",
			__func__, exec_ctl.ccb_exec_status, exec_ctl.nomap_va);
		dax_unlock_pages(dax_ctx, ccb_buf, usr_args.dce_ccb_buf_len);
	}

	/* copyout submitted CCB buf len */
	if (copy_to_user(usr_args.dce_submitted_ccb_buf_len,
			 &exec_ctl.submitted_ccb_buf_len,
			sizeof (exec_ctl.submitted_ccb_buf_len)) != 0) {
		DAX_DBG("%s: copyout dce_submitted_ccb_buf_len "
			"failed, address %p size %ld", __func__,
			usr_args.dce_submitted_ccb_buf_len,
			sizeof (exec_ctl.submitted_ccb_buf_len));
		return -EFAULT;
	}

	/* calculate completion area region offset corresponding to submitted CCBs */
	ca_region_off = ((exec_ctl.submitted_ccb_buf_offset / sizeof (ccb_t))
			 * sizeof (ccb_completion_area_t));

	DAX_DBG("%s: copyout submitted_ccb_buf_len=%d, "
		"ca_region_off=%Ld, ccb_exec_status=%d", __func__,
		exec_ctl.submitted_ccb_buf_len, ca_region_off,
		exec_ctl.ccb_exec_status);

	/* copyout byte offset of completion area region */
	if (copy_to_user(usr_args.dce_ca_region_off, &ca_region_off,
			 sizeof (ca_region_off)) != 0) {
		DAX_DBG("%s: copyout of dce_ca_region_off "
			"failed", __func__);
		return -EFAULT;
	}

	/*copyout dce_nomap_va */
	if (copy_to_user(usr_args.dce_nomap_va, &exec_ctl.nomap_va,
			 sizeof (exec_ctl.nomap_va)) != 0) {
		DAX_DBG("%s: copyout dce_nomap_va failed", __func__);
		return -EFAULT;
	}

	/* copyout dce_ccb_status */
	if (copy_to_user(usr_args.dce_ccb_status, &exec_ctl.ccb_exec_status,
			 sizeof (exec_ctl.ccb_exec_status)) != 0) {
		DAX_DBG("%s: copyout dce_ccb_status failed", __func__);
		return -EFAULT;
	}

	return rv;
}

/*
 * Reserves space in the bip buffer for the user ccbs.  Returns amount reserved
 * which may be less than requested len.
 *
 * If region B exists, then allocate from region B regardless of region A
 * freespace.  Else, compare freespace before and after region A.  If more space
 * before, then create new region B.
 */
static ccb_t *
dax_ccb_buffer_reserve(dax_ctx_t *ctx, size_t len, size_t *reserved)
{
	size_t avail;

	DAX_DBG("%s: len=%ld", __func__, len);

	/* allocate from region B if B exists */
	if (ctx->b_end > 0) {
		avail = ctx->a_start - ctx->b_end;

		if (avail > len)
			avail = len;

		if (avail == 0)
			return (NULL);

		*reserved = avail;
		ctx->resv_start = ctx->b_end;
		ctx->resv_end = ctx->b_end + avail;

		DAX_DBG("%s: region B reserve: reserved=%ld, "
			"resv_start=%d, resv_end=%d, ccb_bufp=0x%p", __func__,
			*reserved, ctx->resv_start, ctx->resv_end,
			(void *)((caddr_t *)(ctx->ccb_buf) + ctx->resv_start));
	} else {

		/*
		 * region A allocation. Check if there is more freespace after
		 * region A than before region A.  Allocate from the larger.
		 */
		avail = ctx->ccb_buflen - ctx->a_end;

		if (avail >= ctx->a_start) {
			/* more freespace after region A */

			if (avail == 0)
				return (NULL);

			if (avail > len)
				avail = len;

			*reserved = avail;
			ctx->resv_start = ctx->a_end;
			ctx->resv_end = ctx->a_end + avail;

			DAX_DBG("%s: region A (after) reserve: "
				"reserved=%ld, resv_start=%d, resv_end=%d, "
				"ccb_bufp=0x%p", __func__, *reserved,
				ctx->resv_start, ctx->resv_end,
				(void *)((caddr_t)(ctx->ccb_buf) +
					 ctx->resv_start));
		} else {
			/* before region A */
			avail = ctx->a_start;

			if (avail == 0)
				return (NULL);

			if (avail > len)
				avail = len;

			*reserved = avail;
			ctx->resv_start = 0;
			ctx->resv_end = avail;

			DAX_DBG("%s: region A (before) reserve: "
				"reserved=%ld, resv_start=%d, resv_end=%d, "
				"ccb_bufp=0x%p", __func__, *reserved,
				ctx->resv_start, ctx->resv_end,
				(void *)((caddr_t)(ctx->ccb_buf) +
					 ctx->resv_start));
		}
	}

	DAX_DBG("%s: a_start=%d a_end=%d, b_end=%d, "
		"resv_start=%d, resv_end=%d, bufcnt=%d", __func__,
		ctx->a_start, ctx->a_end, ctx->b_end,  ctx->resv_start,
		ctx->resv_end, ctx->bufcnt);

	return ((ccb_t *)((caddr_t)(ctx->ccb_buf) + ctx->resv_start));
}

static void
dax_ccb_buffer_commit(dax_ctx_t *ctx, size_t len)
{
	DAX_DBG("%s: len=%ld", __func__, len);

	ASSERT(len <= (ctx->resv_end - ctx->resv_start));

	if (ctx->resv_start == ctx->a_end) {
		ctx->a_end += len;
	} else {
		ctx->b_end += len;
	}

	ctx->resv_start = 0;
	ctx->resv_end = 0;
	ctx->bufcnt += len;

	DAX_DBG("%s: a_start=%d a_end=%d, b_end=%d, "
		"resv_start=%d, resv_end=%d, bufcnt=%d", __func__,
		ctx->a_start, ctx->a_end, ctx->b_end,  ctx->resv_start,
		ctx->resv_end, ctx->bufcnt);
}

/*
 * Return pointer to oldest contig block in buffer, or NULL if empty.
 * In either case, len is set to size of oldest contig block (which may be 0).
 */
static ccb_t *
dax_ccb_buffer_get_contig_block(dax_ctx_t *ctx, size_t *len)
{
	if (ctx->a_end == 0) {
		*len = 0;
		return (NULL);
	} else {
		*len = ctx->a_end - ctx->a_start;
		return ((ccb_t *)((caddr_t)(ctx->ccb_buf) + ctx->a_start));
	}
}

/*
 * Returns amount of contiguous memory decommitted from buffer.
 *
 * Note: If both regions are currently in use, it will only free the
 * memory in region A. If returned amount is less than len, there may
 * be more memory left in buffer.   Caller may need to make multiple
 * calls to decommit all memory in buffer.
 */
static size_t
dax_ccb_buffer_decommit(dax_ctx_t *ctx, size_t len)
{
	size_t a_len;

	DAX_DBG("%s: len=%ld", __func__, len);

	ASSERT(len != 0);
	ASSERT((len % sizeof (ccb_t)) == 0);

	a_len = ctx->a_end - ctx->a_start;

	if (len >= a_len) {
		len = a_len;
		ctx->a_start = 0;
		ctx->a_end = ctx->b_end;
		ctx->b_end = 0;
	} else {
		ctx->a_start += len;
	}

	ctx->bufcnt -= len;

	DAX_DBG("%s: decommited len=%ld, a_start=%d a_end=%d, "
		"b_end=%d, resv_start=%d, resv_end=%d, bufcnt=%d", __func__,
		len, ctx->a_start, ctx->a_end, ctx->b_end,  ctx->resv_start,
		ctx->resv_end, ctx->bufcnt);

	return (len);
}


/*
 * Unlock user pages. Called during dequeue or device close.
 */
static void
dax_unlock_pages(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len)
{
	int i;
	int nelem = ccb_len / sizeof (ccb_t);

	ASSERT(ccb_len % sizeof (ccb_t) == 0);

	for (i = 0; i < nelem; i++) {
		uint32_t idx;

		/* index into ccb_buf */
		idx = &ccb[i] - dax_ctx->ccb_buf;
		if (dax_ctx->pages[idx]) {
			put_page(dax_ctx->pages[idx]);
			DAX_DBG("%s: freeing page %p", __func__, dax_ctx->pages[idx]);
			dax_ctx->pages[idx] = NULL;
		}
	}
}


/*
 * Lock user pages. They get released during the dequeue phase
 * or upon device close.
 */
static int
dax_lock_pages(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len)
{
	int junk, i, ret = 0;
	int nelem = ccb_len / sizeof (ccb_t);
	struct page *page;

	ASSERT(ccb_len % sizeof (ccb_t) == 0);

	for (i = 0; i < nelem; i++) {
		ccb_hdr_t *hdr = CCB_HDR(&ccb[i]);
		uint32_t idx;

		/* index into ccb_buf */
		idx = &ccb[i] - dax_ctx->ccb_buf;

		DAX_DBG("%s: ccb[%d]=0x%p, idx=%d, at_dst=%d",
			__func__, i, &ccb[i], idx, hdr->at_dst);

		/* look at all addresses in hdr, or just output address? */
		if (hdr->at_dst == CCB_AT_VA_ALT) {
			/* what about at_src0, at_src1, at_tbl? */
			/* maybe we only care about at_dst??? */
			unsigned long virtp = ccb[i].dwords[QUERY_DWORD_OUTPUT];
			int res, nr_pages = 1;
			down_read(&current->mm->mmap_sem);
			res = get_user_pages(current, current->mm, virtp, nr_pages, 1, 0, &page, NULL);
			up_read(&current->mm->mmap_sem);

			if (res == nr_pages) {
				//physp =  __pa(page_address(&pages[0]) + (virtp & ~PAGE_MASK));
				dax_ctx->pages[idx] = page;
				DAX_DBG("%s: locked page %p, for VA 0x%lx", __func__, page, virtp);
			} else {
				printk(KERN_WARNING
				       "get_user_pages failed, virtp=0x%lx, nr_pages=%d, res=%d\n", virtp, nr_pages, res);
				ret = -EFAULT;
				break;
			}
		}

#define DAX_REFRESH_TLB(address_type, dword_select) \
		if (hdr->at_##address_type == CCB_AT_VA_ALT && \
		    copy_from_user(&junk, (void *) ccb[i].dwords[QUERY_DWORD_##dword_select], 1)) { \
			DAX_DBG("%s: ccb[%d]=0x%p, idx=%d, bad " #dword_select " address 0x%Lx", \
				__func__, i, &ccb[i], idx, ccb[i].dwords[QUERY_DWORD_##dword_select]); \
		}
		DAX_REFRESH_TLB(dst,  OUTPUT);
		DAX_REFRESH_TLB(src0, INPUT);
		DAX_REFRESH_TLB(src1, SEC_INPUT);
		DAX_REFRESH_TLB(tbl,  TBL);

		if (IS_LONG_CCB(&ccb[i]))
			/* skip over 2nd 64 bytes of long CCB */
			i++;
	}
	if (ret)
		dax_unlock_pages(dax_ctx, ccb, ccb_len);

	return ret;
}

/*
 * Look for use of special dax contiguous segment and
 * set it up for physical access
 */
void
dax_map_segment(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len)
{
	int i;
	int nelem = ccb_len / sizeof (ccb_t);

	ASSERT(ccb_len % sizeof (ccb_t) == 0);
	if (dax_ctx->dax_mm == NULL)
		return;

	for (i = 0; i < nelem; i++) {
		ccb_hdr_t *hdr = CCB_HDR(&ccb[i]);
		uint32_t idx;

		/* index into ccb_buf */
		idx = &ccb[i] - dax_ctx->ccb_buf;

		DAX_DBG("%s: ccb[%d]=0x%p, idx=%d, at_dst=%d",
			__func__, i, &ccb[i], idx, hdr->at_dst);
		/* look at just output address */
		if (hdr->at_dst == CCB_AT_VA_ALT) {
			struct list_head *p;
			dax_vma_t *dv;
			unsigned long virtp = ccb[i].dwords[QUERY_DWORD_OUTPUT];
			ccb_data_acc_ctl_t *access = (ccb_data_acc_ctl_t *) &ccb[i].dwords[QUERY_DWORD_DAC];
			unsigned long output_size = (1 + access->output_buf_sz) * 64;

			DAX_DBG_MAP("%s: dst uva 0x%lx, size=0x%lx", __func__, virtp, output_size);
			spin_lock(&dax_ctx->dax_mm->lock);
			list_for_each(p, &dax_ctx->dax_mm->vma_list) {
				char msg[200];
				dv = list_entry(p, dax_vma_t, vma_list);
				if (dv->vma == NULL) /* cannot use deleted mappings */
					continue;
				if ( (virtp >= dv->vma->vm_start) && (virtp < dv->vma->vm_end)) {
					if ((virtp + output_size - 1) < dv->vma->vm_end) {
						DAX_DBG_MAP("%s: matched %s", __func__, dax_vm_print(msg, dv));
						if (m7_bug & dv->tte) {
							long status;
							hdr->at_dst = CCB_AT_VA;
							ccb[i].dwords[QUERY_DWORD_OUTPUT] = dv->sva + (virtp - dv->vma->vm_start);
							status = mmu_map_addr(dv->sva, 0, dv->tte, HV_MMU_DMMU);
							if (status != 0) {
								DAX_DBG_MAP("%s: mmu_map_addr(0x%lx, 0, 0x%lx, 0x%x) returned 0x%lx",
									    __func__, dv->sva, dv->tte, HV_MMU_DMMU, status);
							} else {
								DAX_DBG_MAP("%s:  changed output to SVA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_OUTPUT]);
							}
						} else {
							hdr->at_dst = CCB_AT_RA;
							ccb[i].dwords[QUERY_DWORD_OUTPUT] = dv->pa + (virtp - dv->vma->vm_start);
							access->flow_ctl = 2; /* enforce flow limit */
							DAX_DBG_MAP("%s:  changed output to RA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_OUTPUT]);
						}
						break;
					}
					else
						DAX_DBG_MAP("%s:  output buffer 0x%lx+0x%lx overflows page 0x%lx+0x%lx", __func__,
							    virtp, output_size, dv->pa, dv->requested_length);
				}
			}
			spin_unlock(&dax_ctx->dax_mm->lock);
		}
		/* then look at source address */
		if (hdr->at_src0 == CCB_AT_VA_ALT) {
			struct list_head *p;
			dax_vma_t *dv;
			unsigned long virtp = ccb[i].dwords[QUERY_DWORD_INPUT];
			ccb_data_acc_ctl_t *access = (ccb_data_acc_ctl_t *) &ccb[i].dwords[QUERY_DWORD_DAC];
			unsigned long input_size = 1 + access->input_cnt;

			DAX_DBG_MAP("%s: src0 uva 0x%lx, size=0x%lx", __func__, virtp, input_size);
			spin_lock(&dax_ctx->dax_mm->lock);
			list_for_each(p, &dax_ctx->dax_mm->vma_list) {
				char msg[200];
				dv = list_entry(p, dax_vma_t, vma_list);
				if (dv->vma == NULL) /* cannot use deleted mappings */
					continue;
				if ( (virtp >= dv->vma->vm_start) && (virtp < dv->vma->vm_end)) {
					if ((virtp + input_size - 1) < dv->vma->vm_end) {
						DAX_DBG_MAP("%s: matched %s", __func__, dax_vm_print(msg, dv));
						if (m7_bug & dv->tte) {
							long status;
							hdr->at_src0 = CCB_AT_VA;
							ccb[i].dwords[QUERY_DWORD_INPUT] = dv->sva + (virtp - dv->vma->vm_start);
							status = mmu_map_addr(dv->sva, 0, dv->tte, HV_MMU_DMMU);
							if (status != 0) {
								DAX_DBG_MAP("%s: mmu_map_addr(0x%lx, 0, 0x%lx, 0x%x) returned 0x%lx",
									    __func__, dv->sva, dv->tte, HV_MMU_DMMU, status);
							} else {
								DAX_DBG_MAP("%s:  changed input to SVA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_INPUT]);
							}
						} else {
							hdr->at_src0 = CCB_AT_RA;
							ccb[i].dwords[QUERY_DWORD_INPUT] = dv->pa + (virtp - dv->vma->vm_start);
							DAX_DBG_MAP("%s:  changed input to RA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_INPUT]);
						}
						break;
					}
					else
						DAX_DBG_MAP("%s:  input buffer 0x%lx+0x%lx overflows page 0x%lx+0x%lx", __func__,
							    virtp, input_size, dv->pa, dv->requested_length);
				}
			}
			spin_unlock(&dax_ctx->dax_mm->lock);
		}
		if (hdr->at_src1 == CCB_AT_VA_ALT) {
			struct list_head *p;
			dax_vma_t *dv;
			unsigned long virtp = ccb[i].dwords[QUERY_DWORD_SEC_INPUT];

			DAX_DBG_MAP("%s: src1 uva 0x%lx", __func__, virtp);
			spin_lock(&dax_ctx->dax_mm->lock);
			list_for_each(p, &dax_ctx->dax_mm->vma_list) {
				char msg[200];
				dv = list_entry(p, dax_vma_t, vma_list);
				if (dv->vma == NULL) /* cannot use deleted mappings */
					continue;
				if ( (virtp >= dv->vma->vm_start) && (virtp < dv->vma->vm_end)) {
					DAX_DBG_MAP("%s: matched %s", __func__, dax_vm_print(msg, dv));
					if (m7_bug & dv->tte) {
						long status;
						hdr->at_src1 = CCB_AT_VA;
						ccb[i].dwords[QUERY_DWORD_SEC_INPUT] = dv->sva + (virtp - dv->vma->vm_start);
						status = mmu_map_addr(dv->sva, 0, dv->tte, HV_MMU_DMMU);
						if (status != 0) {
							DAX_DBG_MAP("%s: mmu_map_addr(0x%lx, 0, 0x%lx, 0x%x) returned 0x%lx",
								    __func__, dv->sva, dv->tte, HV_MMU_DMMU, status);
						} else {
							DAX_DBG_MAP("%s:  changed sec input to SVA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_SEC_INPUT]);
						}
					} else {
						hdr->at_src1 = CCB_AT_RA;
						ccb[i].dwords[QUERY_DWORD_SEC_INPUT] = dv->pa + (virtp - dv->vma->vm_start);
						DAX_DBG_MAP("%s:  changed sec input to RA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_SEC_INPUT]);
					}
					break;
				}
			}
			spin_unlock(&dax_ctx->dax_mm->lock);
		}
		if (hdr->at_tbl == CCB_AT_VA_ALT) {
			struct list_head *p;
			dax_vma_t *dv;
			unsigned long virtp = ccb[i].dwords[QUERY_DWORD_TBL];

			DAX_DBG_MAP("%s: tbl uva 0x%lx", __func__, virtp);
			spin_lock(&dax_ctx->dax_mm->lock);
			list_for_each(p, &dax_ctx->dax_mm->vma_list) {
				char msg[200];
				dv = list_entry(p, dax_vma_t, vma_list);
				if (dv->vma == NULL) /* cannot use deleted mappings */
					continue;
				if ( (virtp >= dv->vma->vm_start) && (virtp < dv->vma->vm_end)) {
					DAX_DBG_MAP("%s: matched %s", __func__, dax_vm_print(msg, dv));
					if (m7_bug & dv->tte) {
						long status;
						hdr->at_tbl = CCB_AT_VA;
						ccb[i].dwords[QUERY_DWORD_TBL] = dv->sva + (virtp - dv->vma->vm_start);
						status = mmu_map_addr(dv->sva, 0, dv->tte, HV_MMU_DMMU);
						if (status != 0) {
							DAX_DBG_MAP("%s: mmu_map_addr(0x%lx, 0, 0x%lx, 0x%x) returned 0x%lx",
								    __func__, dv->sva, dv->tte, HV_MMU_DMMU, status);
						} else {
							DAX_DBG_MAP("%s:  changed tbl to SVA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_TBL]);
						}
					} else {
						hdr->at_tbl = CCB_AT_RA;
						ccb[i].dwords[QUERY_DWORD_TBL] = dv->pa + (virtp - dv->vma->vm_start);
						DAX_DBG_MAP("%s:  changed tbl to RA 0x%Lx", __func__, ccb[i].dwords[QUERY_DWORD_TBL]);
					}
					break;
				}
			}
			spin_unlock(&dax_ctx->dax_mm->lock);
		}
		if (IS_LONG_CCB(&ccb[i]))
			/* skip over 2nd 64 bytes of long CCB */
			i++;
	}

	return;
}

/*
 * Validates user CCB content.  Also sets completion address and address types
 * for all addresses contained in CCB.
 */
static int
dax_preprocess_usr_ccbs(dax_ctx_t *dax_ctx, ccb_t *ccb, size_t ccb_len)
{
	int i;
	int nelem = ccb_len / sizeof (ccb_t);

	ASSERT(ccb_len % sizeof (ccb_t) == 0);

	for (i = 0; i < nelem; i++) {
		ccb_hdr_t *hdr = CCB_HDR(&ccb[i]);
		uint32_t idx;
		ptrdiff_t ca_offset;

		/* enforce validation checks */
		if (dax_validate_ccb(&ccb[i]) != 0) {
			DAX_DBG("%s: ccb[%d] invalid ccb", __func__, i);
			return -ENOKEY;
		}

		/* change all virtual address types to virtual alternate */
		if (hdr->at_src0 == CCB_AT_VA)
			hdr->at_src0 = CCB_AT_VA_ALT;
		if (hdr->at_src1 == CCB_AT_VA)
			hdr->at_src1 = CCB_AT_VA_ALT;
		if (hdr->at_dst == CCB_AT_VA)
			hdr->at_dst = CCB_AT_VA_ALT;
		if (hdr->at_tbl == CCB_AT_VA)
			hdr->at_tbl = CCB_AT_VA_ALT;

		/* set completion (real) address and address type */
		hdr->at_cmpl = CCB_AT_RA;

		idx = &ccb[i] - dax_ctx->ccb_buf;
		ca_offset = (uintptr_t)&dax_ctx->ca_buf[idx] - (uintptr_t)dax_ctx->ca_buf;

		DAX_DBG("%s: ccb[%d]=0x%p, ccb_buf=0x%p, idx=%d "
			"ca_offset=0x%lx, ca_buf_ra=0x%Lx ", __func__, i,
			(void *)&ccb[i], (void *)dax_ctx->ccb_buf, idx,
			ca_offset, dax_ctx->ca_buf_ra);

		DAX_DBG("%s: ccb[%d] setting completion RA=0x%llx",
			__func__, i, dax_ctx->ca_buf_ra + ca_offset);

		CCB_SET_COMPL_PA(dax_ctx->ca_buf_ra + ca_offset,
		    ccb[i].dwords[CCB_DWORD_COMPL]);
		memset((void *)((unsigned long)dax_ctx->ca_buf + ca_offset), 0, sizeof(ccb_completion_area_t));

		if (IS_LONG_CCB(&ccb[i]))
			/* skip over 2nd 64 bytes of long CCB */
			i++;
	}

	return (0);
}

static int
dax_ccb_hv_submit(dax_ctx_t *dax_ctx, ccb_t *ccb_buf, size_t buflen,
		  ccb_exec_ctl_t *exec_ctl)
{
	unsigned long submitted_ccb_buf_len = 0;
	unsigned long nomap_va= 0;
	unsigned long hv_rv;
	int rv = -EIO;
	ptrdiff_t offset;
	long flags = 0;	/* default type_message */

	offset = (uintptr_t)ccb_buf - (uintptr_t)dax_ctx->ccb_buf;

	DAX_DBG("%s: ccb_buf=0x%p, buflen=%ld, offset=0x%lx, "
		"ccb_buf_ra=0x%llx, exec_ctl->hwq_type=%d", __func__,
		(void *)ccb_buf, buflen, offset, dax_ctx->ccb_buf_ra + offset,
		exec_ctl->hwq_type);

	if (dax_debug & DAX_DBG_FLG_BASIC)
		dax_prt_ccbs(ccb_buf, buflen);

	if (exec_ctl->hwq_type == DAX_CCB_HWQ_TYPE_QUERY)
		flags = HV_DAX_QUERY_CMD;
	else
		flags = HV_DAX_MESSAGE_CMD;

	flags |= HV_DAX_CCB_VA_SECONDARY;

	hv_rv = sun4v_dax_ccb_submit((void *) dax_ctx->ccb_buf_ra + offset,
				 buflen, flags, 0,
				 &submitted_ccb_buf_len, &nomap_va);

	exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_INTERNAL;
	exec_ctl->submitted_ccb_buf_len = 0;
	exec_ctl->submitted_ccb_buf_offset = (uint32_t)0;

//	if (hv_rv != HV_EOK)
//		printk(KERN_INFO "%s: failed dax submit, ret=0x%lx\n", __func__, hv_rv);

	DAX_DBG("%s: hcall rv=%ld, submitted_ccb_buf_len=%ld, "
		"nomap_va=0x%lx, flags=0x%lx", __func__, hv_rv, submitted_ccb_buf_len, nomap_va, flags);

	ASSERT(submitted_ccb_buf_len % sizeof (ccb_t) == 0);

	switch (hv_rv) {
	case HV_EOK:
		/*
		 * Hcall succeeded with no errors but the submitted length may
		 * be less than the requested length.  The only way the kernel
		 * can resubmit the remainder is to wait for completion of the
		 * submitted CCBs since there is no way to guarantee the
		 * ordering semantics required by the client applications.
		 * Therefore we let the user library deal with retranmissions.
		 * Ideally we return an error that specifies partial submission
		 * of requested CCBs.
		 *
		 * Note: If HW virtual queues are used, the kernel may be able
		 * to attempt retransmission.
		 */
		rv = 0;
		exec_ctl->ccb_exec_status = DAX_SUBMIT_OK;
		exec_ctl->submitted_ccb_buf_len += submitted_ccb_buf_len;
		exec_ctl->submitted_ccb_buf_offset = (uint32_t)offset;
		break;
	case HV_EWOULDBLOCK:
		/*
		 * This is a transient HV API error that we may eventually want
		 * to hide from the user. For now return EAGAIN and let the user
		 * library retry.
		 */
		DAX_DBG("%s: hcall returned HV_EWOULDBLOCK ", __func__);
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_WOULDBLOCK;
		break;
	case HV_ENOMAP:
		/*
		 * HV was unable to translate a VA.  The VA it could not
		 * translate is returned in the nomap_va param.
		 */
		DAX_DBG("%s: hcall returned HV_ENOMAP "
			"nomap_va=0x%lx", __func__, nomap_va);

		exec_ctl->nomap_va = nomap_va;
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_NOMAP;

		break;
	case HV_EINVAL:
		/*
		 * This is the result of an invalid user CCB as HV is validating
		 * some of the user CCB fields.  Pass this error back to the
		 * user. There is no supporting info to isolate the invalid
		 * field.  May need to enhance the API to tell the user which
		 * field is invalid.
		 */
		DAX_DBG("%s: hcall returned HV_EINVAL ", __func__);
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_CCB_INVAL;
		break;
	case HV_ENOACCESS:
		/*
		 * HV found a VA that did not have the appropriate permissions
		 * (such as the w bit). The VA in question is returned in
		 * nomap_va param, but there is no specific indication which
		 * CCB had the error.  There is no remedy for the kernel to
		 * correct the failure, so return an appropriate error to the
		 * user. Hopefully, something more specific than BadCCB.  Need
		 * to sort out all the error codes when the ccb_exec
		 * implementation is converted to a fast trap handler.
		 */
		DAX_DBG("%s: hcall returned HV_ENOACCESS ", __func__);
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_NOACCESS;
		break;
	case HV_ENOTSUPPORTED:
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_INTERNAL;
		print_err("%s: unsupported operation (%ld)", __func__, hv_rv);
		break;
	default:
		exec_ctl->ccb_exec_status = DAX_SUBMIT_ERR_INTERNAL;
		print_err("%s: unknown hcall return value (%ld)", __func__, hv_rv);
		break;
	}

	return (rv);
}

/*
 * Wait for all CCBs to complete and remove from CCB buffer.
 */
static void
dax_ctx_flush_decommit_ccbs(dax_ctx_t *dax_ctx)
{
	size_t ccb_contig_len;
	size_t decommitted_len;

	DAX_DBG("%s", __func__);

	dax_ctx_flush_ccbs(dax_ctx);

	/* decommit all */
	while (dax_ccb_buffer_get_contig_block(dax_ctx, &ccb_contig_len) != NULL) {
		ASSERT(ccb_contig_len != 0);
		decommitted_len = dax_ccb_buffer_decommit(dax_ctx,
		    ccb_contig_len);
		ASSERT(decommitted_len == ccb_contig_len);
	}
}

/*
 * Wait for all CCBs to complete.  Do not remove from CCB buffer.
 */
static void
dax_ctx_flush_ccbs(dax_ctx_t *dax)
{
	dax_ccb_flush_contig(dax, dax->a_start, dax->a_end);

	if (dax->b_end > 0) {
		dax_ccb_flush_contig(dax, 0, dax->b_end);
	}
}

static void
dax_ccb_flush_contig(dax_ctx_t *dax_ctx, int start_bytes, int end_bytes)
{
	int start_idx = start_bytes / sizeof (ccb_t);
	int end_idx = end_bytes / sizeof (ccb_t);
	int i;

	DAX_DBG("%s: start_idx=%d, end_idx=%d", __func__, start_idx, end_idx);

	for (i = start_idx; i < end_idx; i++) {
		dax_ccb_wait(dax_ctx, i);

		dax_overflow_check(dax_ctx, i);

		/* free any locked page associated with this ccb */
		if (dax_ctx->pages[i]) {
			set_page_dirty(dax_ctx->pages[i]);
			put_page(dax_ctx->pages[i]);
			DAX_DBG("%s: freeing page %p", __func__, dax_ctx->pages[i]);
			dax_ctx->pages[i] = NULL;
		}

		if (IS_LONG_CCB(&dax_ctx->ccb_buf[i])) {
			/* skip over 64B data of long CCB */
			i++;
		}
	}
}

static void
dax_ccb_wait(dax_ctx_t *dax_ctx, int idx)
{
	int nretries = 0;

	DAX_DBG("%s: idx=%d", __func__, idx);

	while (dax_ctx->ca_buf[idx].cmd_status == CCB_CMD_STAT_NOT_COMPLETED) {
		udelay(dax_ccb_wait_usec);

		if (++nretries >= dax_ccb_wait_retries_max) {
			printk(KERN_ALERT "%s: dax_ctx (0x%p): CCB[%d] "
			       "did not complete (timed out, wait usec=%d "
			       "retries=%d). CCB kill will be attempted "
			       "in future version.\n", __func__,
			       (void *)dax_ctx, idx, dax_ccb_wait_usec,
			       dax_ccb_wait_retries_max);
			nretries = 0;
			return;
		}
	}
}

static int dax_address_in_use(dax_vma_t *dv, uint32_t addr_type, unsigned long addr)
{
	if (addr_type == CCB_AT_VA) {
		unsigned long virtp = addr;

		if (virtp >= dv->sva && virtp < dv->sva + dv->actual_length)
			return 1;
	}
	if (addr_type == CCB_AT_RA) {
		unsigned long physp = addr;

		if (physp >= dv->pa && physp < dv->pa + dv->actual_length)
			return 1;
	}

	return 0;
}

#if 0
static int dax_ccb_in_use(dax_ctx_t *ctx, int idx, dax_vma_t *dv)
{
	ccb_t *ccb;
	ccb_hdr_t *hdr;

	if (ctx->ca_buf[idx].cmd_status != CCB_CMD_STAT_NOT_COMPLETED)
		return 0;

	hdr = CCB_HDR(&ctx->ccb_buf[idx]);
	ccb = &ctx->ccb_buf[idx];

	return     dax_address_in_use(dv, hdr->at_dst,  ccb->dwords[QUERY_DWORD_OUTPUT])
		|| dax_address_in_use(dv, hdr->at_src0, ccb->dwords[QUERY_DWORD_INPUT])
		|| dax_address_in_use(dv, hdr->at_src1, ccb->dwords[QUERY_DWORD_SEC_INPUT])
		|| dax_address_in_use(dv, hdr->at_tbl,  ccb->dwords[QUERY_DWORD_TBL]);
}

static int dax_ccbs_in_use_contig(dax_ctx_t *ctx, dax_vma_t *dv, int start_bytes, int end_bytes)
{
	int start_idx = start_bytes / sizeof (ccb_t);
	int end_idx = end_bytes / sizeof (ccb_t);
	int i;

	DAX_DBG("%s: start_idx=%d, end_idx=%d", __func__, start_idx, end_idx);

	for (i = start_idx; i < end_idx; i++) {
		if (dax_ccb_in_use(ctx, i, dv))
			return 1;
		if (IS_LONG_CCB(&ctx->ccb_buf[i])) {
			/* skip over 64B data of long CCB */
			i++;
		}
	}
	return 0;
}


static int
dax_ccbs_in_use(dax_ctx_t *ctx, dax_vma_t *dv)
{
	if (dax_ccbs_in_use_contig(ctx, dv, ctx->a_start, ctx->a_end))
		return 1;
	if (ctx->b_end > 0)
		return dax_ccbs_in_use_contig(ctx, dv, 0, ctx->b_end);
	return 0;
}


static int
dax_vma_in_use(dax_vma_t *dv)
{
	dax_mm_t *dax_mm;
	dax_ctx_t *ctx;
	struct list_head *p;

	/* iterate over all threads in this process, and see if one of them */
	/* has a live reference to this vma */
	dax_mm = dv->dax_mm;
	list_for_each(p, &dax_mm->ctx_list) {
		ctx = list_entry(p, dax_ctx_t, ctx_list);
		if (dax_ccbs_in_use(ctx, dv))
			return 1;
	}
	return 0;
}
#endif

static void
dax_ccb_drain(dax_ctx_t *ctx, int idx, dax_vma_t *dv)
{
	ccb_t *ccb;
	ccb_hdr_t *hdr;

	if (ctx->ca_buf[idx].cmd_status != CCB_CMD_STAT_NOT_COMPLETED)
		return;

	ccb = &ctx->ccb_buf[idx];
	hdr = CCB_HDR(ccb);

	if (dax_address_in_use(dv, hdr->at_dst,  ccb->dwords[QUERY_DWORD_OUTPUT])
	    || dax_address_in_use(dv, hdr->at_src0, ccb->dwords[QUERY_DWORD_INPUT])
	    || dax_address_in_use(dv, hdr->at_src1, ccb->dwords[QUERY_DWORD_SEC_INPUT])
	    || dax_address_in_use(dv, hdr->at_tbl,  ccb->dwords[QUERY_DWORD_TBL]) ) {
		dax_ccb_wait(ctx, idx);
	}
}

static void
dax_ccbs_drain_contig(dax_ctx_t *ctx, dax_vma_t *dv, int start_bytes, int end_bytes)
{
	int start_idx = start_bytes / sizeof (ccb_t);
	int end_idx = end_bytes / sizeof (ccb_t);
	int i;

	DAX_DBG("%s: start_idx=%d, end_idx=%d", __func__, start_idx, end_idx);

	for (i = start_idx; i < end_idx; i++) {
		dax_ccb_drain(ctx, i, dv);
		if (IS_LONG_CCB(&ctx->ccb_buf[i])) {
			/* skip over 64B data of long CCB */
			i++;
		}
	}
}

static void
dax_ccbs_drain(dax_ctx_t *ctx, dax_vma_t *dv)
{
	dax_ccbs_drain_contig(ctx, dv, ctx->a_start, ctx->a_end);
	if (ctx->b_end > 0)
		dax_ccbs_drain_contig(ctx, dv, 0, ctx->b_end);
}

static void
dax_vma_drain(dax_vma_t *dv)
{
	dax_mm_t *dax_mm;
	dax_ctx_t *ctx;
	struct list_head *p;

	/* iterate over all threads in this process and drain all */
	dax_mm = dv->dax_mm;
	list_for_each(p, &dax_mm->ctx_list) {
		ctx = list_entry(p, dax_ctx_t, ctx_list);
		dax_ccbs_drain(ctx, dv);
	}
}

/*
 * Performance Counter Code
 *
 * Author: Dave Aldridge (david.j.aldridge@oracle.com)
 *
 */


/* @brief Write to a performance counter register
 *
 * @param register The register to write to
 * @param value The value to write
 *
 * @return 0 - success
 *         non 0 - failure
 */
static int write_pcr_reg(unsigned long reg, u64 value)
{
	if (!pcr_ops->accessible_pcr(reg)) {
		DAX_DBG_PERF("%s: initial pcr%lu[%016llx]",
			     __func__, reg, pcr_ops->read_pcr(reg));

		pcr_ops->write_pcr(reg, value);
		DAX_DBG_PERF("%s: updated pcr%lu[%016llx]",
			     __func__, reg, pcr_ops->read_pcr(reg));
	} else {
		DAX_DBG_PERF("%s: couldn't access pcr%lu !",
			     __func__, reg);
		return -EACCES;
	}
	return 0;
}


/* @brief Setup the DAX performance counters
 *
 * @param node  The node
 * @param dax   The dax instance
 * @param setup The config value to write
 *
 * @return 0 - success
 *         non 0 - failure
 */
static int dax_setup_counters(unsigned int node, unsigned int dax, u64 setup)
{
	unsigned long reg;

	/* DAX register addresses are given by:
	 *
	 * 171+(node*196)+(dax*4) Performance Counter Control
	 */
	reg = 171 + (node * 196) + (dax * 4);

	return write_pcr_reg(reg, setup);
}

/* @brief Read the DAX performance counters
 *
 * @param node  The node
 * @param dax   The dax instance
 * @param counts Somewhere to write the count values
 *
 * @return 0 - success
 *         non 0 - failure
 */
static int dax_get_counters(unsigned int node, unsigned int dax,
		unsigned long (*counts)[DAX_PER_NODE][COUNTERS_PER_DAX])
{
	int i;
	u64 pcr;
	unsigned long reg;

	/* DAX register addresses are given by:
	 *
	 * 168+(node*196)+(dax*4) Performance Counter 0
	 * 169+(node*196)+(dax*4) Performance Counter 1
	 * 170+(node*196)+(dax*4) Performance Counter 2
	 */
	for (i = 0; i < COUNTERS_PER_DAX; i++) {

		reg = (i + 168) + (node * 196) + (dax * 4);

		if (!pcr_ops->accessible_pcr(reg)) {
			pcr = pcr_ops->read_pcr(reg);
			DAX_DBG_PERF("%s: pcr%lu[%016llx]",
					__func__, reg, pcr);
			counts[node][dax][i] = pcr;
		} else {
			DAX_DBG_PERF("%s: couldn't access pcr%lu !",
					__func__, reg);
			return -EACCES;
		}
	}
	return 0;
}

/* @brief Clear the DAX performance counters
 *
 * @param node  The node
 * @param dax   The dax instance
 *
 * @return 0 - success
 *         non 0 - failure
 */
static int dax_clear_counters(unsigned int node, unsigned int dax)
{
	int ret = 0, i;

	unsigned long reg;

	/* DAX register addresses are given by:
	 *
	 * 168+(node*196)+(dax*4) Performance Counter 0
	 * 169+(node*196)+(dax*4) Performance Counter 1
	 * 170+(node*196)+(dax*4) Performance Counter 2
	 */
	for (i = 0; i < COUNTERS_PER_DAX; i++) {

		reg = (i + 168) + (node * 196) + (dax * 4);
		ret |= write_pcr_reg(reg, 0);
	}
	return ret;
}


static long perfcount_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	unsigned int node, dax;
	unsigned int max_nodes = num_online_nodes();

	unsigned long dax_config;

	/* DAX performance counters are 48 bits wide */
	unsigned long dax_count_bytes =
		max_nodes * DAX_PER_NODE * COUNTERS_PER_DAX * sizeof(u64);

	/* Somwhere to store away the dax performance counter 48 bit values */
	unsigned long (*dax_counts)[DAX_PER_NODE][COUNTERS_PER_DAX];

	switch (cmd) {
	case PERFCOUNT_GET_NODE_COUNT:

		DAX_DBG_PERF("%s: PERFCOUNT_GET_NODE_COUNT: nodes = %u",
			     __func__, max_nodes);

		if (copy_to_user((void *)arg, &max_nodes, sizeof(max_nodes)))
			return -EFAULT;

		return 0;

	case PERFCOUNT_DAX_SET_COUNTERS:

		DAX_DBG_PERF("%s: PERFCOUNT_DAX_SET_COUNTERS", __func__);

		/* Get the performance counter setup from user land */
		if (copy_from_user(&dax_config, (void *)arg, sizeof(unsigned long))) {
			return -EFAULT;
		}

		/* Setup the dax performance counter configuration registers */
		DAX_DBG_PERF("%s: PERFCOUNT_DAX_SET_COUNTERS: dax_config = 0x%lx",
			     __func__, dax_config);

		for (node = 0; node < max_nodes; node++) {
			for (dax = 0; dax < DAX_PER_NODE; dax++) {
				if (dax_setup_counters(node, dax, dax_config) != 0) {
					DAX_DBG_PERF("%s: failed to setup node %u dax %u performance counters",
						__func__, node, dax);
					return -EACCES;
				}
			}
		}

		return 0;

	case PERFCOUNT_DAX_GET_COUNTERS:

		DAX_DBG_PERF("%s: PERFCOUNT_DAX_GET_COUNTERS", __func__);

		/* Somewhere to store the count data */
		dax_counts = kmalloc(dax_count_bytes, GFP_KERNEL);
		if (!dax_counts) {
			ret = -ENOMEM;
			goto dax_get_out;
		}

		/* Read the counters */
		for (node = 0; node < max_nodes; node++) {
			for (dax = 0; dax < DAX_PER_NODE; dax++) {
				if (dax_get_counters(node, dax, dax_counts) != 0) {
					DAX_DBG_PERF("%s: failed to get node %u dax %u performance counters",
						__func__, node, dax);
					ret = -EACCES;
					goto dax_get_err;
				}
			}
		}

		DAX_DBG_PERF("%s: PERFCOUNT_DAX_GET_COUNTERS: copying %lu bytes of perf counter data",
			     __func__, dax_count_bytes);

		if (copy_to_user((void *)arg, dax_counts, dax_count_bytes)) {
			ret = -EFAULT;
		}

dax_get_err:
		kfree(dax_counts);
dax_get_out:
		return ret;

	case PERFCOUNT_DAX_CLEAR_COUNTERS:

		DAX_DBG_PERF("%s: PERFCOUNT_DAX_CLEAR_COUNTERS", __func__);

		/* Clear the counters */
		for (node = 0; node < max_nodes; node++) {
			for (dax = 0; dax < DAX_PER_NODE; dax++) {
				if (dax_clear_counters(node, dax) != 0) {
					DAX_DBG_PERF("%s: failed to clear node %u dax %u performance counters",
						__func__, node, dax);
					ret = -EACCES;
					break;
				}
			}
		}

		return ret;

	default:
		DAX_DBG("%s: Invalid command: 0x%x", __func__, cmd);
		return -ENOTTY;
	}
}

int dax_has_erratum_6008_one_node(void)
{
	ccb_extract_t *ccb;
	ccb_completion_area_t *ca;
	char *mem, *dax_input, *rle_table, *dax_output;

	unsigned long submitted_ccb_buf_len, nomap_va, hv_rv, ra, va;
	unsigned long flow_limit = 0x400, potential_output = 0x1000;
	long timeout;

	mem = kzalloc(0x2000, GFP_KERNEL);
	va = ((unsigned long) mem + 0x7f) & ~0x7fULL; /* align to 128 bytes */

	ccb = (ccb_extract_t *) va;
	ca =  (ccb_completion_area_t *) (va + 0x80);
	dax_input = (char *) va + 0x100;
	rle_table = (char *) va + 0x200;
	memset(rle_table, 0xff, 0x100);
	dax_output = (char *) va + 0x1000;

	ccb->control.hdr.opcode  = CCB_QUERY_OPCODE_EXTRACT;

	/* I/O formats and sizes */
	ccb->control.src0_fmt = CCB_QUERY_IFMT_FIX_BYTE_RLE;
	ccb->control.src0_sz = 0; /* 1 byte */
	ccb->control.src1_enc = CCB_QUERY_SIE_LESS_ONE;
	ccb->control.src1_sz = CCB_QUERY_SZ_EIGHTBIT;
	ccb->control.output_sz = 0;
	ccb->control.output_fmt = CCB_QUERY_OFMT_16B;

	/* addresses */
	*(uint64_t *)&ccb->src0 = (uint64_t) dax_input;
	*(uint64_t *)&ccb->src1 = (uint64_t) rle_table;
	*(uint64_t *)&ccb->output = (uint64_t) dax_output;
	*(uint64_t *)&ccb->completion = (uint64_t) ca;

	/* address types */
	ccb->control.hdr.at_src0 = CCB_AT_VA;
	ccb->control.hdr.at_src1 = CCB_AT_VA;
	ccb->control.hdr.at_dst  = CCB_AT_VA;
	ccb->control.hdr.at_cmpl = CCB_AT_VA;

	/* input sizes and output flow control limit */
	ccb->data_acc_ctl.input_len_fmt = CCB_QUERY_ILF_BYTE;
	/* each byte of input produces a run length of 256 16-byte words */
	ccb->data_acc_ctl.input_cnt = potential_output/(16*256) - 1;
	ccb->data_acc_ctl.output_buf_sz = (flow_limit)/64 - 1;
	ccb->data_acc_ctl.flow_ctl = 2;

	ra = virt_to_phys(ccb);

	hv_rv = sun4v_dax_ccb_submit((void *) ra, 64, HV_DAX_QUERY_CMD, 0, &submitted_ccb_buf_len, &nomap_va);
	if (hv_rv != HV_EOK) {
		printk(KERN_INFO "%s: failed dax submit, ret=0x%lx\n", __func__, hv_rv);
		dax_prt_ccbs((ccb_t *)ccb, 64);
		return 1;
	}

	timeout = 10LL * 1000LL * 1000LL; /* 10ms in ns */
	while (timeout > 0) {
		unsigned long status;
		unsigned long mwait_time = 8192;

		/* monitored load */
		__asm__ __volatile__("lduba [%1] 0x84, %0\n\t"
				     : "=r" (status) : "r" (&ca->cmd_status));
		if (status == CCB_CMD_STAT_NOT_COMPLETED)
			__asm__ __volatile__("wr %0, %%asr28\n\t" /* mwait */
					     : : "r" (mwait_time));
		else
			break;
		timeout = timeout - mwait_time;
	}
	if (timeout <= 0)
		printk(KERN_ALERT "%s: test command timed out\n", __func__);

	hv_rv = ca->output_sz;
	kfree(mem);

	if (hv_rv != flow_limit) {
		printk(KERN_ALERT "%s: 0x%lx bytes output, differs from flow limit 0x%lx\n",
		       __func__, hv_rv, flow_limit);
		printk(KERN_ALERT "%s: mem=%p, va=0x%lx, ccb=%p, ca=%p, out=%p\n",
		       __func__, mem, va, ccb, ca, dax_output);
		return 1;
	}

	return 0;
}
